# Copyright (C) 2024  Loren Eteval <loren.eteval@proton.me>
#
# This file is part of Furious.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

from __future__ import annotations

from Furious.Interface import *
from Furious.PyFramework import *
from Furious.QtFramework import *
from Furious.QtFramework import gettext as _
from Furious.Utility import *
from Furious.Library import *
from Furious.Core import *
from Furious.TrayActions.Import import *
from Furious.Widget.GuiHysteria1 import *
from Furious.Widget.GuiHysteria2 import *
from Furious.Widget.GuiShadowsocks import *
from Furious.Widget.GuiTrojan import *
from Furious.Widget.GuiVLESS import *
from Furious.Widget.GuiVMess import *
from Furious.Window.QRCodeWindow import *
from Furious.Window.TextEditorWindow import *

from PySide6 import QtCore
from PySide6.QtGui import *
from PySide6.QtWidgets import *
from PySide6.QtNetwork import *

from typing import Callable, Union, Sequence, MutableSequence

import queue
import logging
import icmplib
import functools

__all__ = ['UserServersQTableWidget']

logger = logging.getLogger(__name__)

registerAppSettings('ActivatedItemIndex')
registerAppSettings('ServerWidgetSectionSizeTable')

needTrans = functools.partial(needTransFn, source=__name__)


class SubscriptionManager(AppQNetworkAccessManager):
    def __init__(self, parent):
        super().__init__(parent)

        self.networkReplyTable = {}

    def handleFinishedByNetworkReply(self, networkReply):
        assert isinstance(networkReply, QNetworkReply)

        unique, webURL = (
            self.networkReplyTable[networkReply]['unique'],
            self.networkReplyTable[networkReply]['webURL'],
        )

        if networkReply.error() != QNetworkReply.NetworkError.NoError:
            logger.error(f'update subs {webURL} failed. {networkReply.errorString()}')
        else:
            logger.info(f'update subs {webURL} success')

            try:
                data = networkReply.readAll().data()

                uris = list(
                    filter(
                        lambda x: x != '',
                        PyBase64Encoder.decode(data).decode().split('\n'),
                    )
                )
            except Exception as ex:
                # Any non-exit exceptions

                logger.error(f'parse share link failed: {ex}')
            else:
                parent = self.parent()

                if isinstance(parent, UserServersQTableWidget):
                    isConnected = APP().isSystemTrayConnected()

                    subsIndexes = list(
                        index
                        for index, server in enumerate(AS_UserServers())
                        if server.getExtras('subsId') == unique
                    )

                    subsGroupIndex = -1
                    activatedIndex = AS_UserActivatedItemIndex()

                    if activatedIndex in subsIndexes:
                        for index, server in enumerate(AS_UserServers()):
                            if index <= activatedIndex:
                                if server.getExtras('subsId') == unique:
                                    subsGroupIndex += 1
                            else:
                                break

                    parent.deleteItemByIndex(
                        subsIndexes, showTrayMessage=bool(subsGroupIndex < 0)
                    )

                    remaining = len(AS_UserServers())

                    for uri in uris:
                        parent.appendNewItem(config=uri, subsId=unique)

                    if subsGroupIndex >= 0:
                        newIndex = remaining + subsGroupIndex

                        if newIndex < len(AS_UserServers()):
                            parent.activateItemByIndex(newIndex, True)

                            if isConnected and not APP().isSystemTrayConnected():
                                # Trigger connect
                                APP().systemTray.ConnectAction.trigger()

        # Done. Remove entry. Key should be found, but protect it anyway
        self.networkReplyTable.pop(networkReply, None)

    def configureHttpProxy(self, httpProxy: Union[str, None]) -> bool:
        useProxy = super().configureHttpProxy(httpProxy)

        if useProxy:
            logger.info(f'update subs uses proxy server {httpProxy}')
        else:
            logger.info(f'update subs uses no proxy')

        return useProxy

    def updateSubs(self, unique: str, webURL: str):
        networkReply = self.get(QNetworkRequest(QtCore.QUrl(webURL)))

        self.networkReplyTable[networkReply] = {
            'unique': unique,
            'webURL': webURL,
        }

        networkReply.finished.connect(
            functools.partial(
                self.handleFinishedByNetworkReply,
                networkReply,
            )
        )


class WorkerSequence(ItemUpdateProtocol):
    def __init__(self, sequence: Sequence, index: int, item: ConfigurationFactory):
        super().__init__(sequence, index, item)

    def currentItemDeleted(self) -> bool:
        return super().currentItemDeleted() or FastItemDeletionSearch.isInTrash(
            self.currentItem
        )

    def updateResult(self):
        # Extra guard
        if APP() is None or APP().isExiting():
            return

        super().updateResult()


class TestPingLatencyWorker(WorkerSequence, QtCore.QObject, QtCore.QRunnable):
    finished = QtCore.Signal(int, object)

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        # Explictly called __init__
        QtCore.QObject.__init__(self)
        QtCore.QRunnable.__init__(self)

    def updateImpl(self):
        self.finished.emit(self.currentIndex, self.currentItem)

    def run(self):
        if self.currentItemDeleted():
            # Deleted. Do nothing
            return

        assert isinstance(self.currentItem, ConfigurationFactory)

        try:
            result = icmplib.ping(
                self.currentItem.itemAddress,
                count=1,
                timeout=2,
                interval=1,
            )
        except Exception as ex:
            # Any non-exit exceptions

            def classname(ob) -> str:
                return ob.__class__.__name__

            self.currentItem.setExtras('delayResult', classname(ex))
            self.updateResult()
        else:
            # Result address should not be empty
            if result.address and result.is_alive:
                self.currentItem.setExtras('delayResult', f'{round(result.avg_rtt)}ms')
            else:
                if result.packet_loss == 1:
                    self.currentItem.setExtras('delayResult', 'Timeout')
                else:
                    self.currentItem.setExtras('delayResult', 'Error')

            self.updateResult()


class TestTcpingLatencyWorker(WorkerSequence, QtCore.QObject, QtCore.QRunnable):
    finished = QtCore.Signal(int, object)

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        # Explictly called __init__
        QtCore.QObject.__init__(self)
        QtCore.QRunnable.__init__(self)

    def updateImpl(self):
        self.finished.emit(self.currentIndex, self.currentItem)

    def run(self):
        if self.currentItemDeleted():
            # Deleted. Do nothing
            return

        assert isinstance(self.currentItem, ConfigurationFactory)

        try:
            sent, rtts = tcping(
                self.currentItem.itemAddress,
                int(self.currentItem.itemPort.split(',')[0]),
                count=1,
                timeout=2,
                interval=1,
            )
        except Exception as ex:
            # Any non-exit exceptions

            def classname(ob) -> str:
                return ob.__class__.__name__

            self.currentItem.setExtras('delayResult', classname(ex))
            self.updateResult()
        else:
            if rtts:
                self.currentItem.setExtras('delayResult', f'{round(rtts[0] * 1000)}ms')
            else:
                self.currentItem.setExtras('delayResult', 'Timeout')

            self.updateResult()


class TestDownloadSpeedWorker(WorkerSequence, QtCore.QObject):
    progressed = QtCore.Signal(int, object)

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        # Explictly called __init__
        QtCore.QObject.__init__(self)

        self.hasSpeedResult = False
        self.totalBytesRead = 0

        self.coreManager = CoreManager()

        self.networkAccessManager = QNetworkAccessManager()
        self.networkReply = None
        self.elapsedTimer = QtCore.QElapsedTimer()

    def isFinished(self) -> bool:
        if isinstance(self.networkReply, QNetworkReply):
            return self.networkReply.isFinished()
        else:
            return True

    def abort(self):
        if isinstance(self.networkReply, QNetworkReply):
            self.networkReply.abort()

    def updateImpl(self):
        self.progressed.emit(self.currentIndex, self.currentItem)

    def run(self):
        if self.currentItemDeleted():
            # Deleted. Do nothing
            return

        assert isinstance(self.currentItem, ConfigurationFactory)

        if not self.currentItem.isValid():
            self.currentItem.setExtras('speedResult', 'Invalid')
            self.updateResult()

            return

        def coreExitCallback(config: ConfigurationFactory, exitcode: int):
            if exitcode == CoreProcess.ExitCode.ConfigurationError:
                self.currentItem.setExtras('speedResult', 'Invalid')

                return self.updateResult()
            if exitcode == CoreProcess.ExitCode.ServerStartFailure:
                self.currentItem.setExtras('speedResult', 'Start failed')

                return self.updateResult()
            if exitcode == CoreProcess.ExitCode.SystemShuttingDown:
                pass
            else:
                self.currentItem.setExtras('speedResult', f'Core exited {exitcode}')

                return self.updateResult()

        def msgCallback(line: str):
            try:
                APP().logViewerWindowCore.appendLine(line)
            except Exception:
                # Any non-exit exceptions

                pass

        self.currentItem.setExtras('speedResult', 'Starting')
        self.updateResult()

        copy = self.currentItem.deepcopy()

        if isinstance(copy, ConfigurationXray):
            # Force redirect
            copy['inbounds'] = [
                {
                    'tag': 'http',
                    'port': 20809,
                    'listen': '127.0.0.1',
                    'protocol': 'http',
                    'sniffing': {
                        'enabled': True,
                        'destOverride': [
                            'http',
                            'tls',
                        ],
                    },
                    'settings': {
                        'auth': 'noauth',
                        'udp': True,
                        'allowTransparent': False,
                    },
                },
            ]

            try:
                for outboundObject in copy['outbounds']:
                    if outboundObject['tag'] == 'proxy':
                        # Avoid confusion with potentially existing 'proxy' tag
                        outboundObject['tag'] = 'proxy20809'
            except Exception:
                # Any non-exit exceptions

                pass

            self.coreManager.start(
                copy,
                'Global',
                coreExitCallback,
                msgCallback=msgCallback,
                deepcopy=False,
                proxyModeOnly=True,
                log=False,
            )
        elif isinstance(copy, ConfigurationHysteria1) or isinstance(
            copy, ConfigurationHysteria2
        ):
            # Force redirect
            copy['http'] = {
                'listen': '127.0.0.1:20809',
                'timeout': 300,
                'disable_udp': False,
            }

            # No socks inbounds
            copy.pop('socks5', '')

            self.coreManager.start(
                copy,
                'Global',
                coreExitCallback,
                msgCallback=msgCallback,
                deepcopy=False,
                proxyModeOnly=True,
                log=False,
            )
        else:
            self.currentItem.setExtras('speedResult', 'Invalid')
            self.updateResult()

            return

        self.networkAccessManager.setProxy(
            QNetworkProxy(QNetworkProxy.ProxyType.HttpProxy, '127.0.0.1', 20809)
        )
        self.networkReply = self.networkAccessManager.get(
            QNetworkRequest(
                QtCore.QUrl(
                    'http://speed.cloudflare.com/__down?during=download&bytes=104857600'
                )
            )
        )
        self.networkReply.readyRead.connect(self.handleReadyRead)
        self.networkReply.finished.connect(self.handleFinished)
        self.elapsedTimer.start()

    @QtCore.Slot()
    def handleReadyRead(self):
        if self.coreManager.allRunning():
            self.totalBytesRead += self.networkReply.readAll().length()

            # Convert to seconds
            elapsedSecond = self.elapsedTimer.elapsed() / 1000
            downloadSpeed = self.totalBytesRead / elapsedSecond / 1024 / 1024

            # Has speed test result
            self.hasSpeedResult = True
            self.currentItem.setExtras('speedResult', f'{downloadSpeed:.2f} M/s')
            self.updateResult()

    @QtCore.Slot()
    def handleFinished(self):
        if self.networkReply.error() != QNetworkReply.NetworkError.NoError:
            if not self.hasSpeedResult:
                if not self.coreManager.allRunning():
                    # Core ExitCallback has been called
                    return

                if (
                    self.networkReply.error()
                    == QNetworkReply.NetworkError.OperationCanceledError
                ):
                    # Canceled by application
                    self.currentItem.setExtras('speedResult', 'Canceled')
                else:
                    try:
                        error = self.networkReply.error().name
                    except Exception:
                        # Any non-exit exceptions

                        error = 'Unknown Error'

                    if isinstance(error, bytes):
                        # Some old version PySide6 returns it as bytes. Protect it.
                        errorString = error.decode('utf-8', 'replace')
                    elif isinstance(error, str):
                        errorString = error
                    else:
                        errorString = 'Unknown Error'

                    if errorString.endswith('Error'):
                        self.currentItem.setExtras('speedResult', errorString[:-5])
                    else:
                        self.currentItem.setExtras('speedResult', errorString)
        else:
            if self.coreManager.allRunning():
                self.totalBytesRead += self.networkReply.readAll().length()

                # Convert to seconds
                elapsedSecond = self.elapsedTimer.elapsed() / 1000
                downloadSpeed = self.totalBytesRead / elapsedSecond / 1024 / 1024

                self.currentItem.setExtras('speedResult', f'{downloadSpeed:.2f} M/s')
            else:
                self.currentItem.setExtras('speedResult', 'Start failed')

        self.coreManager.stopAll()
        self.updateResult()


class UserServersQTableWidgetHorizontalHeader(AppQHeaderView):
    class SortOrder:
        Ascending_ = False
        Descending = True

    @staticmethod
    def emptyClickGuard():
        return False

    def __init__(self, *args, **kwargs):
        self.clickGuardFn = kwargs.pop('clickGuardFn', self.emptyClickGuard)
        self.customSortFn = kwargs.pop('customSortFn', None)

        super().__init__(QtCore.Qt.Orientation.Horizontal, *args, **kwargs)

        self.sortOrderTable = list(
            self.SortOrder.Ascending_ for i in range(self.columnCount)
        )

        self.sectionClicked.connect(self.handleSectionClicked)

    def customSort(self, clickedIndex):
        self.customSortFn(clickedIndex, reverse=self.sortOrderTable[clickedIndex])
        # Toggle
        self.sortOrderTable[clickedIndex] = not self.sortOrderTable[clickedIndex]

    @QtCore.Slot(int)
    def handleSectionClicked(self, clickedIndex: int):
        if callable(self.clickGuardFn) and self.clickGuardFn():
            # Guarded. Do nothing
            return

        if self.customSortFn is None:
            # Sorting not supported
            return

        parent = self.parent()

        if isinstance(parent, AppQTableWidget):
            # Support item activation
            activatedIndex = AS_UserActivatedItemIndex()

            if activatedIndex < 0:
                activatedServerId = None
            else:
                activatedServerId = id(AS_UserServers()[activatedIndex])

                # De-activated temporarily for sorting
                parent.activateItemByIndex(activatedIndex, False)

            self.customSort(clickedIndex)

            if activatedServerId is not None:
                foundActivatedItem = False

                for index, server in enumerate(AS_UserServers()):
                    if activatedServerId == id(server):
                        foundActivatedItem = True

                        parent.setCurrentItem(parent.item(index, 0))
                        parent.activateItemByIndex(index, True)

                        break

                # Object id should be found
                assert foundActivatedItem

        else:
            # Not yet implemented
            pass


class UserServersQTableWidgetVerticalHeader(AppQHeaderView):
    def __init__(self, *args, **kwargs):
        super().__init__(QtCore.Qt.Orientation.Vertical, *args, **kwargs)


class UserServersQTableWidgetHeaders:
    def __init__(self, name: str, func: Callable[[ConfigurationFactory], str] = None):
        self.name = name
        self.func = func

    def __call__(self, item: ConfigurationFactory) -> str:
        if callable(self.func):
            return self.func(item)
        else:
            return getattr(item, f'item{self}')

    def __eq__(self, other):
        return str(self) == str(other)

    def __str__(self):
        return self.name


needTrans(
    'Remark',
    'Protocol',
    'Address',
    'Port',
    'Transport',
    'TLS',
    'Subscription',
    'Latency',
    'Speed',
    'Customize JSON Configuration...',
    'Move Up',
    'Move Down',
    'Duplicate',
    'Delete',
    'Select All',
    'Scroll To Activated Server',
    'Test Ping Latency',
    'Test Tcping Latency',
    'Test Download Speed',
    'Clear Test Results',
    'New Empty Configuration',
    'Export Share Link To Clipboard',
    'Export As QR Code',
    'Export JSON Configuration To Clipboard',
    'Connecting',
    'Connecting. Please wait',
    'Untitled',
)


class UserServersQTableWidget(QTranslatable, AppQTableWidget):
    Headers = [
        UserServersQTableWidgetHeaders('Remark'),
        UserServersQTableWidgetHeaders('Protocol'),
        UserServersQTableWidgetHeaders('Address'),
        UserServersQTableWidgetHeaders('Port'),
        UserServersQTableWidgetHeaders('Transport'),
        UserServersQTableWidgetHeaders('TLS'),
        UserServersQTableWidgetHeaders('Subscription'),
        UserServersQTableWidgetHeaders('Latency'),
        UserServersQTableWidgetHeaders('Speed'),
    ]

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.subsManager = SubscriptionManager(parent=self)

        self.testDownloadSpeedQueue = queue.Queue()
        self.testDownloadSpeedTimer = QtCore.QTimer()
        self.testDownloadSpeedTimer.timeout.connect(self.handleTestDownloadSpeedJob)
        self.testDownloadSpeedTimer.start(250)

        # Text Editor Window
        self.textEditorWindow = TextEditorWindow()

        # Must set before flush all
        self.setColumnCount(len(self.Headers))

        # Flush all data to table
        self.flushAll()

        # Install custom header
        self.setHorizontalHeader(
            UserServersQTableWidgetHorizontalHeader(
                parent=self,
                clickGuardFn=lambda: False,
                customSortFn=self.customSortFn,
                sectionSizeSettingsName='ServerWidgetSectionSizeTable',
            )
        )
        self.setVerticalHeader(UserServersQTableWidgetVerticalHeader(self))

        self.horizontalHeader().setCustomSectionResizeMode()
        self.horizontalHeader().restoreSectionSize()

        self.setHorizontalHeaderLabels(list(_(str(header)) for header in self.Headers))

        # Selection
        self.setSelectionColor(AppHue.disconnectedColor())
        self.setSelectionBehavior(QTableWidget.SelectionBehavior.SelectRows)
        self.setSelectionMode(QTableWidget.SelectionMode.ExtendedSelection)

        # No drag and drop
        self.setDragEnabled(False)
        self.setDragDropMode(QAbstractItemView.DragDropMode.NoDragDrop)
        self.setDropIndicatorShown(False)
        self.setDefaultDropAction(QtCore.Qt.DropAction.IgnoreAction)

        self.editConfigActionRef = AppQAction(
            _('Customize JSON Configuration...'),
            icon=bootstrapIcon('pencil-square.svg'),
            callback=lambda: self.editSelectedItemConfiguration(),
            shortcut=QtCore.QKeyCombination(
                QtCore.Qt.KeyboardModifier.ControlModifier,
                QtCore.Qt.Key.Key_E,
            ),
        )

        contextMenuActions = [
            AppQAction(
                _('Move Up'),
                callback=lambda: self.moveUpSelectedItem(),
            ),
            AppQAction(
                _('Move Down'),
                callback=lambda: self.moveDownSelectedItem(),
            ),
            AppQAction(
                _('Duplicate'),
                callback=lambda: self.duplicateSelectedItem(),
            ),
            AppQAction(
                _('Delete'),
                callback=lambda: self.deleteSelectedItem(),
            ),
            AppQSeperator(),
            self.editConfigActionRef,
            AppQSeperator(),
            AppQAction(
                _('Select All'),
                callback=lambda: self.selectAll(),
                shortcut=QtCore.QKeyCombination(
                    QtCore.Qt.KeyboardModifier.ControlModifier,
                    QtCore.Qt.Key.Key_A,
                ),
            ),
            AppQSeperator(),
            AppQAction(
                _('Scroll To Activated Server'),
                callback=lambda: self.scrollToActivatedItem(),
                shortcut=QtCore.QKeyCombination(
                    QtCore.Qt.KeyboardModifier.ControlModifier,
                    QtCore.Qt.Key.Key_G,
                ),
            ),
            AppQSeperator(),
            AppQAction(
                _('Test Ping Latency'),
                callback=lambda: self.testSelectedItemPingLatency(),
                shortcut=QtCore.QKeyCombination(
                    QtCore.Qt.KeyboardModifier.ControlModifier,
                    QtCore.Qt.Key.Key_P,
                ),
            ),
            AppQAction(
                _('Test Tcping Latency'),
                callback=lambda: self.testSelectedItemTcpingLatency(),
                shortcut=QtCore.QKeyCombination(
                    QtCore.Qt.KeyboardModifier.ControlModifier,
                    QtCore.Qt.Key.Key_O,
                ),
            ),
            AppQAction(
                _('Test Download Speed'),
                callback=lambda: self.testSelectedItemDownloadSpeed(),
                shortcut=QtCore.QKeyCombination(
                    QtCore.Qt.KeyboardModifier.ControlModifier,
                    QtCore.Qt.Key.Key_T,
                ),
            ),
            AppQAction(
                _('Clear Test Results'),
                callback=lambda: self.clearSelectedItemTestResult(),
                shortcut=QtCore.QKeyCombination(
                    QtCore.Qt.KeyboardModifier.ControlModifier,
                    QtCore.Qt.Key.Key_R,
                ),
            ),
            AppQSeperator(),
            AppQAction(
                _('New Empty Configuration'),
                callback=lambda: self.newEmptyItem(),
                shortcut=QtCore.QKeyCombination(
                    QtCore.Qt.KeyboardModifier.ControlModifier,
                    QtCore.Qt.Key.Key_N,
                ),
            ),
            ImportFromFileAction(),
            ImportURIFromClipboardAction(
                shortcut=QtCore.QKeyCombination(
                    QtCore.Qt.KeyboardModifier.ControlModifier,
                    QtCore.Qt.Key.Key_V,
                ),
            ),
            ImportJSONFromClipboardAction(
                shortcut=QtCore.QKeyCombination(
                    QtCore.Qt.KeyboardModifier.ControlModifier
                    | QtCore.Qt.KeyboardModifier.ShiftModifier,
                    QtCore.Qt.Key.Key_J,
                ),
            ),
            AppQSeperator(),
            AppQAction(
                _('Export Share Link To Clipboard'),
                callback=lambda: self.exportSelectedItemURI(),
                shortcut=QtCore.QKeyCombination(
                    QtCore.Qt.KeyboardModifier.ControlModifier,
                    QtCore.Qt.Key.Key_C,
                ),
            ),
            AppQAction(
                _('Export As QR Code'),
                icon=bootstrapIcon('qr-code.svg'),
                callback=lambda: self.exportSelectedItemQR(),
            ),
            AppQAction(
                _('Export JSON Configuration To Clipboard'),
                callback=lambda: self.exportSelectedItemJSON(),
                shortcut=QtCore.QKeyCombination(
                    QtCore.Qt.KeyboardModifier.ControlModifier,
                    QtCore.Qt.Key.Key_J,
                ),
            ),
        ]

        self.contextMenu = AppQMenu(*contextMenuActions)

        # Add actions to self in order to activate shortcuts
        self.addActions(self.contextMenu.actions())
        self.setContextMenuPolicy(QtCore.Qt.ContextMenuPolicy.CustomContextMenu)
        self.customContextMenuRequested.connect(self.handleCustomContextMenuRequested)

        # Distinguish double-click and activated
        self.doubleClickedFlag = False

        # Signals
        self.itemChanged.connect(self.handleItemChanged)
        self.itemSelectionChanged.connect(self.handleItemSelectionChanged)
        self.itemActivated.connect(self.handleItemActivated)
        self.itemDoubleClicked.connect(self.handleItemDoubleClicked)

        if self.activatedItem() is not None:
            self.setCurrentItem(self.activatedItem())
            self.activateItemByIndex(AS_UserActivatedItemIndex(), True)

    @QtCore.Slot(QTableWidgetItem)
    def handleItemChanged(self, item: QTableWidgetItem):
        index = item.row()

        if self.item(index, 0) is None:
            return

        itemText = self.item(index, 0).text()

        if AS_UserServers()[index].getExtras('remark') != itemText:
            AS_UserServers()[index].setExtras('remark', itemText)

    @QtCore.Slot()
    def handleItemSelectionChanged(self):
        if len(self.selectedIndex) > 1:
            self.editConfigActionRef.setDisabled(True)
        else:
            self.editConfigActionRef.setDisabled(False)

    @QtCore.Slot(QTableWidgetItem)
    def handleItemActivated(self, item: QTableWidgetItem):
        if self.doubleClickedFlag:
            # Ignore double-click
            self.doubleClickedFlag = False

            return

        oldIndex = AS_UserActivatedItemIndex()
        newIndex = item.row()

        if oldIndex == newIndex:
            # Same item activated. Do nothing
            return

        if APP().systemTray.ConnectAction.isConnecting():
            mbox = AppQMessageBox(icon=AppQMessageBox.Icon.Information)
            mbox.setWindowTitle(_('Connecting'))
            mbox.setText(_('Connecting. Please wait'))

            if PLATFORM != 'Darwin':
                # Show the MessageBox asynchronously
                mbox.open()
            else:
                # Show the MessageBox asynchronously
                # TODO: Verify
                mbox.open()

            return

        if oldIndex >= 0:
            self.activateItemByIndex(oldIndex, False)

        self.activateItemByIndex(newIndex, True)

        if APP().isSystemTrayConnected():
            APP().systemTray.ConnectAction.doDisconnect()
            APP().systemTray.ConnectAction.trigger()

    @functools.lru_cache(None)
    def getGuiEditorByProtocol(self, protocol, **kwargs):
        logger.debug(f'getGuiEditorByProtocol called with protocol {protocol}')

        if protocolRepr(protocol) == Protocol.VMess:
            return GuiVMess(parent=self, **kwargs)
        if protocolRepr(protocol) == Protocol.VLESS:
            return GuiVLESS(parent=self, **kwargs)
        if protocolRepr(protocol) == Protocol.Shadowsocks:
            return GuiShadowsocks(parent=self, **kwargs)
        if protocolRepr(protocol) == Protocol.Trojan:
            return GuiTrojan(parent=self, **kwargs)
        if protocolRepr(protocol) == Protocol.Hysteria2:
            return GuiHysteria2(parent=self, **kwargs)
        if protocolRepr(protocol) == Protocol.Hysteria1:
            return GuiHysteria1(parent=self, **kwargs)

        return None

    def getGuiEditorByFactory(self, factory, **kwargs):
        if isinstance(factory, ConfigurationXray):
            protocol = factory.proxyProtocol

            guiEditor = self.getGuiEditorByProtocol(protocol, **kwargs)
        elif isinstance(factory, ConfigurationHysteria2):
            guiEditor = self.getGuiEditorByProtocol(Protocol.Hysteria2, **kwargs)
        elif isinstance(factory, ConfigurationHysteria1):
            guiEditor = self.getGuiEditorByProtocol(Protocol.Hysteria1, **kwargs)
        else:
            guiEditor = None

        return guiEditor

    @QtCore.Slot(QTableWidgetItem)
    def handleItemDoubleClicked(self, item: QTableWidgetItem):
        self.doubleClickedFlag = True

        index = item.row()
        factory = AS_UserServers()[index]

        guiEditor = self.getGuiEditorByFactory(factory, translatable=False)

        if guiEditor is None:
            # Unrecognized. Do nothing
            return

        guiEditor.setWindowTitle(f'{index + 1} - ' + factory.getExtras('remark'))

        try:
            guiEditor.factoryToInput(factory)
        except Exception as ex:
            # Any non-exit exceptions

            logger.error(f'error while converting factory to input: {ex}')

        guiEditor.accepted.connect(
            functools.partial(
                self.handleGuiEditorAccepted,
                guiEditor,
                index,
                factory,
            )
        )
        guiEditor.rejected.connect(
            functools.partial(
                self.handleGuiEditorRejected,
                guiEditor,
            )
        )
        guiEditor.open()

    def handleGuiEditorAccepted(
        self,
        editor: GuiEditorWidgetQDialog,
        index: int,
        factory: ConfigurationFactory,
    ):
        logger.debug(f'guiEditor accepted with index {index}')

        modified = editor.inputToFactory(factory)

        # Still flush to row since remark may be modified
        self.flushRow(index, factory)

        if modified and index == AS_UserActivatedItemIndex():
            try:
                if APP().isSystemTrayConnected():
                    mbox = NewChangesNextTimeMBox()

                    # Show the MessageBox asynchronously
                    mbox.open()
            except Exception:
                # Any non-exit exceptions

                pass

        editor.accepted.disconnect()
        editor.rejected.disconnect()

    def handleGuiEditorRejected(self, editor: GuiEditorWidgetQDialog):
        editor.accepted.disconnect()
        editor.rejected.disconnect()

    @QtCore.Slot(QtCore.QPoint)
    def handleCustomContextMenuRequested(self, point):
        self.contextMenu.exec(self.mapToGlobal(point))

    def customSortFn(self, clickedIndex, **kwargs):
        def keyFn(server):
            data = self.Headers[clickedIndex](server)

            if clickedIndex == self.Headers.index('Latency'):
                if data.endswith('ms'):
                    # Strip value
                    data = data[:-2]

                try:
                    return float(data)
                except Exception:
                    # Any non-exit exceptions

                    return abs(hash(data)) + 2**20

            if clickedIndex == self.Headers.index('Speed'):
                if data.endswith(' M/s'):
                    # Strip value
                    data = data[:-4]

                try:
                    return float(data)
                except Exception:
                    # Any non-exit exceptions

                    return abs(hash(data)) + 2**20

            return data

        AS_UserServers().sort(key=keyFn, **kwargs)

        self.flushAll()

    def activatedItem(self) -> QTableWidgetItem:
        return self.item(AS_UserActivatedItemIndex(), 0)

    def activateItemByIndex(self, index, activate):
        super().activateItemByIndex(index, activate)

        if activate:
            AppSettings.set('ActivatedItemIndex', str(index))

    def flushItem(self, row: int, column: int, item: ConfigurationFactory):
        header = self.Headers[column]

        oldItem = self.item(row, column)
        newItem = QTableWidgetItem(header(item))

        if oldItem is None:
            # Item does not exists
            newItem.setFont(QFont(APP().customFontName))

            if str(header) == 'Latency' or str(header) == 'Speed':
                # Test results. Align right and vcenter
                newItem.setTextAlignment(
                    QtCore.Qt.AlignmentFlag.AlignRight
                    | QtCore.Qt.AlignmentFlag.AlignVCenter
                )
        else:
            # Use existing
            newItem.setFont(oldItem.font())
            newItem.setForeground(oldItem.foreground())

            if oldItem.textAlignment() != 0:
                newItem.setTextAlignment(oldItem.textAlignment())

        #
        # Legacy flags. Not used
        #
        # if str(header) == 'Remark':
        #     # Remark is editable
        #     newItem.setFlags(
        #         QtCore.Qt.ItemFlag.ItemIsEnabled
        #         | QtCore.Qt.ItemFlag.ItemIsSelectable
        #         | QtCore.Qt.ItemFlag.ItemIsEditable
        #     )
        # else:
        #     newItem.setFlags(
        #         QtCore.Qt.ItemFlag.ItemIsEnabled | QtCore.Qt.ItemFlag.ItemIsSelectable
        #     )

        # Remark is now editable via GUI window
        newItem.setFlags(
            QtCore.Qt.ItemFlag.ItemIsEnabled | QtCore.Qt.ItemFlag.ItemIsSelectable
        )

        self.setItem(row, column, newItem)

    def addServerViaGui(
        self,
        protocol: str,
        windowTitle: str = APPLICATION_NAME,
        **kwargs,
    ):
        factory = getEmptyFactory(protocol)

        guiEditor = self.getGuiEditorByFactory(factory, **kwargs)

        if guiEditor is None:
            # Unrecognized. Do nothing
            return

        guiEditor.setWindowTitle(windowTitle)

        try:
            guiEditor.factoryToInput(factory)
        except Exception as ex:
            # Any non-exit exceptions

            logger.error(f'error while converting factory to input: {ex}')

        guiEditor.accepted.connect(
            functools.partial(
                self.handleAddServerViaGuiAccepted,
                guiEditor,
                factory,
            )
        )
        guiEditor.rejected.connect(
            functools.partial(
                self.handleAddServerViaGuiRejected,
                guiEditor,
            )
        )
        guiEditor.open()

    def handleAddServerViaGuiAccepted(
        self,
        editor: GuiEditorWidgetQDialog,
        factory: ConfigurationFactory,
    ):
        editor.inputToFactory(factory)

        self.appendNewItemByFactory(factory)

        editor.accepted.disconnect()
        editor.rejected.disconnect()

    def handleAddServerViaGuiRejected(self, editor: GuiEditorWidgetQDialog):
        editor.accepted.disconnect()
        editor.rejected.disconnect()

    def flushRow(self, row: int, item: ConfigurationFactory):
        for column in list(range(self.columnCount())):
            self.flushItem(row, column, item)

    def flushAll(self):
        if self.rowCount() == 0:
            # Should insert row
            for index, item in enumerate(AS_UserServers()):
                self.insertRow(index)
                self.flushRow(index, item)
        else:
            for index, item in enumerate(AS_UserServers()):
                self.flushRow(index, item)

    def swapItem(self, index0: int, index1: int):
        def swapSequenceItem(sequence: MutableSequence, param0: int, param1: int):
            swap = sequence[param0]

            sequence[param0] = sequence[param1]
            sequence[param1] = swap

        swapSequenceItem(AS_UserServers(), index0, index1)

        self.flushRow(index0, AS_UserServers()[index0])
        self.flushRow(index1, AS_UserServers()[index1])

        if index0 == AS_UserActivatedItemIndex():
            # De-activate
            self.activateItemByIndex(index0, False)
            # Activate
            self.activateItemByIndex(index1, True)
        elif index1 == AS_UserActivatedItemIndex():
            # De-activate
            self.activateItemByIndex(index1, False)
            # Activate
            self.activateItemByIndex(index0, True)

    def newEmptyItem(self):
        self.appendNewItem(remark=_('Untitled'), acceptInvalid=True)

    def moveUpItemByIndex(self, index):
        if index <= 0 or index >= len(AS_UserServers()):
            # The top item, or does not exist. Do nothing
            return

        self.swapItem(index, index - 1)

    def moveUpSelectedItem(self):
        indexes = self.selectedIndex

        if len(indexes) == 0:
            # Nothing selected. Do nothing
            return

        for index in indexes:
            self.moveUpItemByIndex(index)

        with QBlockSignals(self):
            self.setCurrentIndex(self.indexFromItem(self.item(indexes[-1] - 1, 0)))

        self.selectMultipleRows(list(index - 1 for index in indexes), True)

    def moveDownItemByIndex(self, index):
        if index < 0 or index >= len(AS_UserServers()) - 1:
            # The bottom item, or does not exist. Do nothing
            return

        self.swapItem(index, index + 1)

    def moveDownSelectedItem(self):
        indexes = self.selectedIndex

        if len(indexes) == 0:
            # Nothing selected. Do nothing
            return

        for index in indexes[::-1]:
            self.moveDownItemByIndex(index)

        with QBlockSignals(self):
            self.setCurrentIndex(self.indexFromItem(self.item(indexes[0] + 1, 0)))

        self.selectMultipleRows(list(index + 1 for index in indexes), True)

    def duplicateSelectedItem(self):
        indexes = self.selectedIndex

        if len(indexes) == 0:
            # Nothing selected. Do nothing
            return

        for index in indexes:
            if 0 <= index < len(AS_UserServers()):
                # Do not clone subsId
                self.appendNewItem(
                    remark=AS_UserServers()[index].getExtras('remark'),
                    config=AS_UserServers()[index],
                )

    def deleteItemByIndex(self, indexes, showTrayMessage=True) -> int:
        if len(indexes) == 0:
            # Nothing selected. Do nothing
            return 0

        if AS_UserActivatedItemIndex() in indexes:
            deleteActivated = True
        else:
            deleteActivated = False

        for i in range(len(indexes)):
            deleteIndex = indexes[i] - i

            with QBlockSignals(self):
                self.removeRow(deleteIndex)

            FastItemDeletionSearch.moveToTrash(AS_UserServers()[deleteIndex])

            AS_UserServers().pop(deleteIndex)

            if not deleteActivated and deleteIndex < AS_UserActivatedItemIndex():
                AppSettings.set(
                    'ActivatedItemIndex', str(AS_UserActivatedItemIndex() - 1)
                )

        if deleteActivated:
            # Set invalid first
            AppSettings.set('ActivatedItemIndex', str(-1))

            if APP().isSystemTrayConnected():
                if showTrayMessage:
                    # Trigger disconnect
                    APP().systemTray.ConnectAction.trigger()
                else:
                    # Trigger disconnect silently
                    APP().systemTray.ConnectAction.doDisconnect()

        return len(indexes)

    def deleteSelectedItem(self):
        indexes = self.selectedIndex

        if len(indexes) == 0:
            # Nothing selected. Do nothing
            return

        def handleResultCode(_indexes, code):
            if code == PySide6LegacyEnumValueWrapper(AppQMessageBox.StandardButton.Yes):
                self.deleteItemByIndex(_indexes)
            else:
                pass

        mbox = QuestionDeleteMBox(icon=AppQMessageBox.Icon.Question)
        mbox.isMulti = bool(len(indexes) > 1)
        mbox.possibleRemark = f'{indexes[0] + 1} - {self.item(indexes[0], 0).text()}'
        mbox.setText(mbox.customText())
        mbox.finished.connect(functools.partial(handleResultCode, indexes))

        # dummy ref
        setattr(self, '_questionDeleteMBox', mbox)

        # Show the MessageBox asynchronously
        mbox.open()

    def editSelectedItemConfiguration(self):
        indexes = self.selectedIndex

        if len(indexes) == 0:
            # Nothing selected. Do nothing
            return

        if len(indexes) != 1:
            # Should not reach here
            return

        index = indexes[0]
        title = f'{index + 1} - ' + AS_UserServers()[index].getExtras('remark')

        self.textEditorWindow.currentIndex = index
        self.textEditorWindow.customWindowTitle = title
        self.textEditorWindow.setWindowTitle(title)
        self.textEditorWindow.setPlainText(AS_UserServers()[index].toJSONString(), True)
        self.textEditorWindow.show()

    def scrollToActivatedItem(self):
        activatedItem = self.activatedItem()

        self.setCurrentItem(activatedItem)
        self.scrollToItem(activatedItem)

    def testSelectedItemPingLatency(self):
        indexes = self.selectedIndex

        if len(indexes) == 0:
            # Nothing selected. Do nothing
            return

        # Real selected factory
        references = list(AS_UserServers()[index] for index in indexes)

        for index, reference in zip(indexes, references):
            worker = TestPingLatencyWorker(AS_UserServers(), index, reference)

            worker.setAutoDelete(True)
            worker.finished.connect(
                lambda paramIndex, paramFactory: self.flushItem(
                    paramIndex, self.Headers.index('Latency'), paramFactory
                )
            )

            APP().threadPool.start(worker)

    def testSelectedItemTcpingLatency(self):
        indexes = self.selectedIndex

        if len(indexes) == 0:
            # Nothing selected. Do nothing
            return

        # Real selected factory
        references = list(AS_UserServers()[index] for index in indexes)

        for index, reference in zip(indexes, references):
            worker = TestTcpingLatencyWorker(AS_UserServers(), index, reference)

            worker.setAutoDelete(True)
            worker.finished.connect(
                lambda paramIndex, paramFactory: self.flushItem(
                    paramIndex, self.Headers.index('Latency'), paramFactory
                )
            )

            APP().threadPool.start(worker)

    @QtCore.Slot()
    def handleTestDownloadSpeedJob(self):
        try:
            index, server = self.testDownloadSpeedQueue.get_nowait()
        except queue.Empty:
            # Queue is empty

            return

        def testDownloadSpeed(counter=0, timeout=5000, step=100):
            worker = TestDownloadSpeedWorker(AS_UserServers(), index, server)
            worker.progressed.connect(
                lambda paramIndex, paramFactory: self.flushItem(
                    paramIndex, self.Headers.index('Speed'), paramFactory
                )
            )
            worker.run()

            while (
                not APP().isExiting() and not worker.isFinished() and counter < timeout
            ):
                PySide6LegacyEventLoopWait(step)

                counter += step

            if not worker.isFinished():
                worker.abort()

                while not worker.isFinished():
                    PySide6LegacyEventLoopWait(step)

            if APP().isExiting():
                # Stop timer
                self.testDownloadSpeedTimer.stop()

        testDownloadSpeed()

    def testSelectedItemDownloadSpeed(self):
        indexes = self.selectedIndex

        if len(indexes) == 0:
            # Nothing selected. Do nothing
            return

        # Real selected factory
        references = list(AS_UserServers()[index] for index in indexes)

        for index, reference in zip(indexes, references):
            try:
                # Served by FIFO
                self.testDownloadSpeedQueue.put_nowait((index, reference))
            except Exception:
                # Any non-exit exceptions

                pass

    def clearSelectedItemTestResult(self):
        indexes = self.selectedIndex

        if len(indexes) == 0:
            # Nothing selected. Do nothing
            return

        for index in indexes:
            factory = AS_UserServers()[index]
            factory.setExtras('delayResult', '')
            factory.setExtras('speedResult', '')

            self.flushItem(index, self.Headers.index('Latency'), factory)
            self.flushItem(index, self.Headers.index('Speed'), factory)

    def updateSubs(self, httpProxy: Union[str, None]):
        self.subsManager.configureHttpProxy(httpProxy)

        for key, value in AS_UserSubscription().items():
            self.subsManager.updateSubs(key, value.get('webURL', ''))

    def appendNewItemByFactory(self, factory: ConfigurationFactory):
        AS_UserServers().append(factory)

        row = self.rowCount()

        self.insertRow(row)
        self.flushRow(row, factory)

        if len(AS_UserServers()) == 1:
            # The first one. Click it
            self.setCurrentItem(self.item(0, 0))

            # Try to be user-friendly in some extreme cases
            if not APP().isSystemTrayConnected():
                # Activate automatically
                self.activateItemByIndex(0, True)

    def appendNewItem(self, **kwargs):
        acceptInvalid = kwargs.pop('acceptInvalid', False)

        model = {
            'remark': kwargs.pop('remark', ''),
            'config': kwargs.pop('config', ''),
            'subsId': kwargs.pop('subsId', ''),
        }
        tostr = f'{model}'

        factory = constructFromAny(model.pop('config', ''), **model)

        if factory.isValid():
            self.appendNewItemByFactory(factory)
        else:
            if acceptInvalid:
                self.appendNewItemByFactory(factory)
            else:
                logger.error(f'invalid item: {tostr}')

    def exportSelectedItemURI(self):
        indexes = self.selectedIndex

        if len(indexes) == 0:
            # Nothing selected. Do nothing
            return

        def toURI(factory) -> str:
            assert isinstance(factory, ConfigurationFactory)

            try:
                return factory.toURI()
            except Exception:
                # Any non-exit exceptions

                return ''

        # TODO: MessageBox?
        QApplication.clipboard().setText(
            '\n'.join(list(toURI(AS_UserServers()[index]) for index in indexes))
        )

    def exportSelectedItemQR(self):
        indexes = self.selectedIndex

        if len(indexes) == 0:
            # Nothing selected. Do nothing
            return

        window = QRCodeWindow()
        window.initTabByIndex(indexes)

        if window.tabCount() > 0:
            window.show()

    def exportSelectedItemJSON(self):
        indexes = self.selectedIndex

        if len(indexes) == 0:
            # Nothing selected. Do nothing
            return

        # TODO: MessageBox?
        QApplication.clipboard().setText(
            '\n'.join(list(AS_UserServers()[index].toJSONString() for index in indexes))
        )

    def showTabAndSpaces(self):
        self.textEditorWindow.showTabAndSpaces()

    def hideTabAndSpaces(self):
        self.textEditorWindow.hideTabAndSpaces()

    def keyPressEvent(self, event):
        if event.key() == QtCore.Qt.Key.Key_Delete:
            self.deleteSelectedItem()
        elif event.key() == QtCore.Qt.Key.Key_Return:
            if PLATFORM == 'Darwin':
                # Activate by Enter key on macOS
                self.itemActivated.emit(self.currentItem())
            else:
                super().keyPressEvent(event)
        else:
            super().keyPressEvent(event)

    def disconnectedCallback(self):
        super().disconnectedCallback()

        self.activateItemByIndex(AS_UserActivatedItemIndex(), True)

    def connectedCallback(self):
        super().connectedCallback()

        self.activateItemByIndex(AS_UserActivatedItemIndex(), True)

    def retranslate(self):
        self.setHorizontalHeaderLabels(list(_(str(header)) for header in self.Headers))



# Copyright (C) 2024  Loren Eteval <loren.eteval@proton.me>
#
# This file is part of Furious.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

from __future__ import annotations

from Furious.Utility.Constants import *
from Furious.Utility.Utility import parseHostPort, runExternalCommand
from Furious.Utility.AppSettings import *

import logging
import subprocess

__all__ = ['SystemProxy']

logger = logging.getLogger(__name__)


def handleAppSystemProxyMode() -> bool:
    try:
        if AppSettings.get('SystemProxyMode') == 'Auto':
            # Automatically configure
            return True
        else:
            # Do not change
            return False
    except Exception:
        # Any non-exit exceptions

        # Automatically configure
        return True


def linuxProxyConfig(proxy_args, arg0, arg1):
    runExternalCommand(
        [
            'gsettings',
            'set',
            'org.gnome.system.' + proxy_args,
            arg0,
            arg1,
        ],
        stdout=subprocess.DEVNULL,
        stderr=subprocess.DEVNULL,
        check=True,
    )


def darwinProxyConfig(operation, *args):
    def getNetworkServices():
        command = runExternalCommand(
            ['networksetup', '-listallnetworkservices'],
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            check=True,
        )

        # Replace with command.stdout.decode('utf-8', 'replace')...?
        service = list(filter(lambda x: x != '', command.stdout.decode().split('\n')))

        return service[1:]

    for serviceName in getNetworkServices():
        runExternalCommand(
            [
                'networksetup',
                f'-{operation}',
                serviceName,
                *args,
            ]
        )


class _SystemProxy:
    def __init__(self):
        self._daemonThread = None

    @staticmethod
    def pac(pac_url):
        def _pac():
            if PLATFORM == 'Windows':
                try:
                    import sysproxy

                    return sysproxy.pac(pac_url)
                except Exception:
                    # Any non-exit exceptions

                    return False

            if PLATFORM == 'Linux':
                try:
                    linuxProxyConfig('proxy', 'autoconfig-url', pac_url)
                    linuxProxyConfig('proxy', 'mode', 'auto')
                except Exception:
                    # Any non-exit exceptions

                    return False
                else:
                    return True

            if PLATFORM == 'Darwin':
                try:
                    darwinProxyConfig('setautoproxyurl', pac_url)
                except Exception:
                    # Any non-exit exceptions

                    return False
                else:
                    return True

        if not handleAppSystemProxyMode():
            logger.info(f'ignore proxy PAC \'{pac_url}\' request')

            return

        if _pac():
            logger.info(f'set proxy PAC \'{pac_url}\' success')
        else:
            logger.error(f'set proxy PAC \'{pac_url}\' failed')

    @staticmethod
    def set(server, bypass):
        def _set():
            if PLATFORM == 'Windows':
                try:
                    import sysproxy

                    return sysproxy.set(server, bypass)
                except Exception:
                    # Any non-exit exceptions

                    return False

            if PLATFORM == 'Linux':
                try:
                    host, port = parseHostPort(server)

                    linuxProxyConfig('proxy.http', 'host', host)
                    linuxProxyConfig('proxy.http', 'port', port)
                    linuxProxyConfig('proxy.https', 'host', host)
                    linuxProxyConfig('proxy.https', 'port', port)
                    linuxProxyConfig(
                        'proxy', 'ignore-hosts', str(list(bypass.split(';')))
                    )
                    linuxProxyConfig('proxy', 'mode', 'manual')
                except Exception:
                    # Any non-exit exceptions

                    return False
                else:
                    return True

            if PLATFORM == 'Darwin':
                try:
                    darwinProxyConfig('setwebproxy', *parseHostPort(server))
                    darwinProxyConfig('setsecurewebproxy', *parseHostPort(server))
                    darwinProxyConfig('setproxybypassdomains', *bypass.split(';'))
                except Exception:
                    # Any non-exit exceptions

                    return False
                else:
                    return True

        if not handleAppSystemProxyMode():
            logger.info(f'ignore proxy server {server} request')

            return

        if _set():
            logger.info(f'set proxy server {server} success')
        else:
            logger.error(f'set proxy server {server} failed')

    @staticmethod
    def off():
        def _off():
            if PLATFORM == 'Windows':
                try:
                    import sysproxy

                    return sysproxy.off()
                except Exception:
                    # Any non-exit exceptions

                    return False

            if PLATFORM == 'Linux':
                try:
                    linuxProxyConfig('proxy', 'mode', 'none')
                except Exception:
                    # Any non-exit exceptions

                    return False
                else:
                    return True

            if PLATFORM == 'Darwin':
                try:
                    darwinProxyConfig('setwebproxystate', 'off')
                    darwinProxyConfig('setsecurewebproxystate', 'off')
                    darwinProxyConfig('setautoproxystate', 'off')
                except Exception:
                    # Any non-exit exceptions

                    return False
                else:
                    return True

        if not handleAppSystemProxyMode():
            logger.info('ignore turn off proxy request')

            return

        if _off():
            logger.info('turn off proxy success')
        else:
            logger.error('turn off proxy failed')

    def daemonOn_(self):
        def _daemonOn_():
            if PLATFORM == 'Windows':
                try:
                    import sysproxy
                    import threading

                    if self._daemonThread is not None:
                        return False

                    self._daemonThread = threading.Thread(
                        target=lambda: sysproxy.daemon_on_(), daemon=True
                    )
                    self._daemonThread.start()

                    return True
                except Exception:
                    # Any non-exit exceptions

                    return False

        if not handleAppSystemProxyMode():
            logger.info('ignore turn on proxy daemon request')

            return

        if _daemonOn_():
            logger.info('turn on proxy daemon success')
        else:
            logger.error('turn on proxy daemon failed')

    def daemonOff(self):
        def _daemonOff():
            if PLATFORM == 'Windows':
                try:
                    import sysproxy
                    import threading

                    if self._daemonThread is None:
                        # Already in off state
                        return True

                    assert isinstance(self._daemonThread, threading.Thread)

                    if sysproxy.daemon_off():
                        self._daemonThread.join()
                        # Reset it
                        self._daemonThread = None

                        return True
                    else:
                        return False
                except Exception:
                    # Any non-exit exceptions

                    return False

            # Note: try to resolve non-Windows proxy daemon issue by
            # turning on StartupOnBoot by default. This should be
            # friendly for most of the users

        if not handleAppSystemProxyMode():
            logger.info('ignore turn off proxy daemon request')

            return

        if _daemonOff():
            logger.info('turn off proxy daemon success')
        else:
            logger.error('turn off proxy daemon failed')


SystemProxy = _SystemProxy()



# Copyright (C) 2024  Loren Eteval <loren.eteval@proton.me>
#
# This file is part of Furious.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

from __future__ import annotations

from Furious.Interface import *
from Furious.PyFramework import *
from Furious.QtFramework import *
from Furious.QtFramework import gettext as _
from Furious.Utility import *
from Furious.Library import *

from PySide6 import QtCore
from PySide6.QtGui import *
from PySide6.QtWidgets import *

from typing import Callable

import functools

__all__ = ['UserSubsQTableWidget']

registerAppSettings('SubscriptionWidgetSectionSizeTable')

needTrans = functools.partial(needTransFn, source=__name__)


class UserSubsQTableWidgetHorizontalHeader(AppQHeaderView):
    def __init__(self, *args, **kwargs):
        super().__init__(QtCore.Qt.Orientation.Horizontal, *args, **kwargs)


class UserSubsQTableWidgetVerticalHeader(AppQHeaderView):
    def __init__(self, *args, **kwargs):
        super().__init__(QtCore.Qt.Orientation.Vertical, *args, **kwargs)


class UserSubsQTableWidgetHeaders:
    def __init__(self, name: str, func: Callable[[dict], str] = None):
        self.name = name
        self.func = func

    def __call__(self, item: dict) -> str:
        if callable(self.func):
            return self.func(item)
        else:
            return ''

    def __eq__(self, other):
        return str(self) == str(other)

    def __str__(self):
        return self.name


needTrans(
    'Remark',
    'Delete',
)


class UserSubsQTableWidget(QTranslatable, AppQTableWidget):
    Headers = [
        UserSubsQTableWidgetHeaders('Remark', lambda item: item.get('remark', '')),
        UserSubsQTableWidgetHeaders('URL', lambda item: item.get('webURL', '')),
    ]

    def __init__(self, *args, **kwargs):
        self.deleteUniqueCallback = kwargs.pop('deleteUniqueCallback', None)

        super().__init__(*args, **kwargs)

        # Must set before flush all
        self.setColumnCount(len(self.Headers))

        # Flush all data to table
        self.flushAll()

        # Install custom header
        self.setHorizontalHeader(
            UserSubsQTableWidgetHorizontalHeader(
                parent=self,
                sectionSizeSettingsName='SubscriptionWidgetSectionSizeTable',
            )
        )
        self.setVerticalHeader(UserSubsQTableWidgetVerticalHeader(self))

        self.horizontalHeader().setCustomSectionResizeMode()
        self.horizontalHeader().restoreSectionSize()

        self.setHorizontalHeaderLabels(list(_(str(header)) for header in self.Headers))

        # Selection
        self.setSelectionColor(AppHue.disconnectedColor())
        self.setSelectionBehavior(QTableWidget.SelectionBehavior.SelectRows)
        self.setSelectionMode(QTableWidget.SelectionMode.ExtendedSelection)

        # No drag and drop
        self.setDragEnabled(False)
        self.setDragDropMode(QAbstractItemView.DragDropMode.NoDragDrop)
        self.setDropIndicatorShown(False)
        self.setDefaultDropAction(QtCore.Qt.DropAction.IgnoreAction)

        contextMenuActions = [
            AppQAction(
                _('Delete'),
                callback=lambda: self.deleteSelectedItem(),
            ),
        ]

        self.contextMenu = AppQMenu(*contextMenuActions)
        self.setContextMenuPolicy(QtCore.Qt.ContextMenuPolicy.CustomContextMenu)

        # Signals
        self.itemChanged.connect(self.handleItemChanged)
        self.customContextMenuRequested.connect(self.handleCustomContextMenuRequested)

    @QtCore.Slot(QTableWidgetItem)
    def handleItemChanged(self, item: QTableWidgetItem):
        unique = list(AS_UserSubscription().keys())[item.row()]
        keyMap = ['remark', 'webURL']

        AS_UserSubscription()[unique][keyMap[item.column()]] = item.text()

    @QtCore.Slot(QtCore.QPoint)
    def handleCustomContextMenuRequested(self, point):
        self.contextMenu.exec(self.mapToGlobal(point))

    def deleteSelectedItem(self):
        indexes = self.selectedIndex

        if len(indexes) == 0:
            # Nothing to do
            return

        def handleResultCode(_indexes, code):
            if code == PySide6LegacyEnumValueWrapper(AppQMessageBox.StandardButton.Yes):
                for i in range(len(_indexes)):
                    deleteIndex = _indexes[i] - i
                    deleteUnique = list(AS_UserSubscription().keys())[deleteIndex]

                    self.removeRow(deleteIndex)

                    AS_UserSubscription().pop(deleteUnique)

                    if callable(self.deleteUniqueCallback):
                        self.deleteUniqueCallback(deleteUnique)
            else:
                # Do not delete
                pass

        mbox = QuestionDeleteMBox(icon=AppQMessageBox.Icon.Question)
        mbox.isMulti = bool(len(indexes) > 1)
        mbox.possibleRemark = self.item(indexes[0], 0).text()
        mbox.setText(mbox.customText())
        mbox.finished.connect(functools.partial(handleResultCode, indexes))

        # dummy ref
        setattr(self, '_questionDeleteMBox', mbox)

        # Show the MessageBox asynchronously
        mbox.open()

    def flushItem(self, row, column, item):
        header = self.Headers[column]

        oldItem = self.item(row, column)
        newItem = QTableWidgetItem(header(item))

        if oldItem is None:
            # Item does not exists
            newItem.setFont(QFont(APP().customFontName))
        else:
            # Use existing
            newItem.setFont(oldItem.font())
            newItem.setForeground(oldItem.foreground())

            if oldItem.textAlignment() != 0:
                newItem.setTextAlignment(oldItem.textAlignment())

            # Editable
            newItem.setFlags(
                QtCore.Qt.ItemFlag.ItemIsEnabled
                | QtCore.Qt.ItemFlag.ItemIsSelectable
                | QtCore.Qt.ItemFlag.ItemIsEditable
            )

        self.setItem(row, column, newItem)

    def flushRow(self, row, item):
        for column in list(range(self.columnCount())):
            self.flushItem(row, column, item)

    def flushAll(self):
        if self.rowCount() == 0:
            # Should insert row
            for index, key in enumerate(AS_UserSubscription()):
                self.insertRow(index)
                self.flushRow(index, AS_UserSubscription()[key])
        else:
            for index, key in enumerate(AS_UserSubscription()):
                self.flushRow(index, AS_UserSubscription()[key])

    def appendNewItem(self, **kwargs):
        unique = kwargs.pop('unique', '')
        remark = kwargs.pop('remark', '')
        webURL = kwargs.pop('webURL', '')

        # The internal subs object
        subs = {
            unique: {
                'remark': remark,
                'webURL': webURL,
            }
        }

        AS_UserSubscription().update(subs)

        row = self.rowCount()

        self.insertRow(row)
        self.flushRow(row, subs[unique])

    def retranslate(self):
        self.setHorizontalHeaderLabels(list(_(str(header)) for header in self.Headers))



# Copyright (C) 2024  Loren Eteval <loren.eteval@proton.me>
#
# This file is part of Furious.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

from __future__ import annotations

from Furious.Interface import *
from Furious.QtFramework.DynamicTranslate import gettext as _, needTransFn
from Furious.QtFramework.QtWidgets import *
from Furious.Utility import *

from PySide6 import QtCore
from PySide6.QtWidgets import *

from typing import Callable, Sequence

__all__ = [
    'GuiEditorItemTextInput',
    'GuiEditorItemTextSpinBox',
    'GuiEditorItemTextComboBox',
    'GuiEditorItemTextCheckBox',
    'GuiEditorItemBasicRemark',
    'GuiEditorItemProxyHttp',
    'GuiEditorItemProxySocks',
    'GuiEditorWidgetQWidget',
    'GuiEditorWidgetQGroupBox',
    'GuiEditorWidgetQDialog',
]

needTrans = functools.partial(needTransFn, source=__name__)


class GuiEditorItemTextInput(GuiEditorItemWidgetContainer):
    def __init__(self, *args, **kwargs):
        title = kwargs.pop('title', '')
        translatable = kwargs.pop('translatable', True)
        parent = kwargs.pop('parent', None)

        super().__init__(*args, **kwargs)

        self._title = AppQLabel(_(title), translatable=translatable, parent=parent)
        self._input = QLineEdit(parent=parent)

    def text(self) -> str:
        return self._input.text()

    def setText(self, text: str):
        self._input.setText(text)

    def widgets(self):
        return self._title, self._input


class GuiEditorItemTextSpinBox(GuiEditorItemWidgetContainer):
    def __init__(self, *args, **kwargs):
        title = kwargs.pop('title', '')
        translatable = kwargs.pop('translatable', True)
        parent = kwargs.pop('parent', None)

        super().__init__(*args, **kwargs)

        self._title = AppQLabel(_(title), translatable=translatable, parent=parent)
        self._input = QSpinBox(parent=parent)

    def value(self) -> int:
        return self._input.value()

    def setValue(self, value: int):
        self._input.setValue(value)

    def setRange(self, minRange: int, maxRange: int):
        self._input.setRange(minRange, maxRange)

    def widgets(self):
        return self._title, self._input


class GuiEditorItemTextComboBox(GuiEditorItemWidgetContainer):
    def __init__(self, *args, **kwargs):
        title = kwargs.pop('title', '')
        translatable = kwargs.pop('translatable', True)
        parent = kwargs.pop('parent', None)

        super().__init__(*args, **kwargs)

        self._title = AppQLabel(_(title), translatable=translatable, parent=parent)
        self._input = QComboBox(parent=parent)

    def text(self) -> str:
        return self._input.currentText()

    def setText(self, text: str):
        self._input.setCurrentText(text)

    def addItems(self, texts: Sequence[str]):
        self._input.addItems(texts)

    def connectActivated(self, func: Callable):
        self._input.activated.connect(func)

    def widgets(self):
        return self._title, self._input


class GuiEditorItemTextCheckBox(GuiEditorItemWidgetContainer):
    def __init__(self, *args, **kwargs):
        title = kwargs.pop('title', '')
        translatable = kwargs.pop('translatable', True)
        parent = kwargs.pop('parent', None)

        super().__init__(*args, **kwargs)

        self._input = AppQCheckBox(_(title), translatable=translatable, parent=parent)

    def isChecked(self) -> bool:
        return self._input.isChecked()

    def setChecked(self, checked: bool):
        self._input.setChecked(checked)

    def widgets(self):
        return (self._input,)


class GuiEditorItemBasicRemark(GuiEditorItemTextInput):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def inputToFactory(self, config: ConfigurationFactory) -> bool:
        oldRemark = config.getExtras('remark')
        newRemark = self.text()

        if newRemark != oldRemark:
            config.setExtras('remark', newRemark)

            # Value modified, but not return as a modified behavior
            return False
        else:
            # Not modified
            return False

    def factoryToInput(self, config: ConfigurationFactory):
        self.setText(config.getExtras('remark'))


class GuiEditorItemProxyHttp(GuiEditorItemTextInput):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def inputToFactory(self, config: ConfigurationFactory) -> bool:
        oldHttp = config.httpProxyEndpoint()
        newHttp = self.text()

        if newHttp == '':
            if oldHttp == '':
                return False
            else:
                config.setHttpProxyEndpoint(newHttp)

                return True
        else:
            if isinstance(oldHttp, str):
                if newHttp != oldHttp:
                    config.setHttpProxyEndpoint(newHttp)

                    return True
                else:
                    return False
            else:
                config.setHttpProxyEndpoint(newHttp)

                return True

    def factoryToInput(self, config: ConfigurationFactory):
        try:
            self.setText(config.httpProxyEndpoint())
        except Exception:
            # Any non-exit exceptions

            self.setText('')


class GuiEditorItemProxySocks(GuiEditorItemTextInput):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def inputToFactory(self, config: ConfigurationFactory) -> bool:
        oldSocks = config.socksProxyEndpoint()
        newSocks = self.text()

        if newSocks == '':
            if oldSocks == '':
                return False
            else:
                config.setSocksProxyEndpoint(newSocks)

                return True
        else:
            if isinstance(oldSocks, str):
                if newSocks != oldSocks:
                    config.setSocksProxyEndpoint(newSocks)

                    return True
                else:
                    return False
            else:
                config.setSocksProxyEndpoint(newSocks)

                return True

    def factoryToInput(self, config: ConfigurationFactory):
        try:
            self.setText(config.socksProxyEndpoint())
        except Exception:
            # Any non-exit exceptions

            self.setText('')


class GuiEditorWidget(GuiEditorItemFactory):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self._containers = self.containerSequence()

    def containerSequence(self) -> Sequence[GuiEditorItemWidgetContainer]:
        raise NotImplementedError

    def inputToFactory(self, config: ConfigurationFactory) -> bool:
        modified = False

        for container in self._containers:
            try:
                modified |= container.inputToFactory(config)
            except Exception:
                # Any non-exit exceptions

                modified |= False

        return modified

    def factoryToInput(self, config: ConfigurationFactory):
        for container in self._containers:
            try:
                container.factoryToInput(config)
            except Exception:
                # Any non-exit exceptions

                pass


class GuiEditorWidgetQWidget(GuiEditorWidget, QWidget):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        layout = QFormLayout()
        layout.setFormAlignment(QtCore.Qt.AlignmentFlag.AlignLeft)
        layout.setFieldGrowthPolicy(QFormLayout.FieldGrowthPolicy.ExpandingFieldsGrow)

        for container in self._containers:
            layout.addRow(*container.widgets())

        self.setLayout(layout)


class GuiEditorWidgetQGroupBox(GuiEditorWidget, AppQGroupBox):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        formLayout = QFormLayout()
        formLayout.setFormAlignment(QtCore.Qt.AlignmentFlag.AlignLeft)
        formLayout.setFieldGrowthPolicy(
            QFormLayout.FieldGrowthPolicy.ExpandingFieldsGrow
        )

        for container in self._containers:
            formLayout.addRow(*container.widgets())

        page = QWidget()
        page.setLayout(formLayout)

        # Align
        self._widget = QStackedWidget()
        self._widget.addWidget(page)

        vboxLayout = QVBoxLayout()
        vboxLayout.addWidget(self._widget)

        self.setLayout(vboxLayout)

        # layout = QFormLayout()
        #
        # for container in self._containers:
        #     layout.addRow(*container.widgets())
        #
        # self.setLayout(layout)


needTrans('Cancel')


class GuiEditorWidgetQDialog(GuiEditorItemFactory, AppQDialog):
    def __init__(self, *args, **kwargs):
        tabText = kwargs.pop('tabText', '')
        tabTranslatable = kwargs.pop('tabTranslatable', False)

        super().__init__(*args, **kwargs)

        self.setWindowModality(QtCore.Qt.WindowModality.WindowModal)
        self.setFixedSize(
            int(450 * GOLDEN_RATIO * GOLDEN_RATIO), int(450 * GOLDEN_RATIO)
        )

        self.tabCentralWidget = QWidget()
        self.tabCentralWidgetLayout = QGridLayout(self.tabCentralWidget)

        self.groupBoxes = self.groupBoxSequence()

        for index, groupBox in enumerate(self.groupBoxes):
            self.tabCentralWidgetLayout.addWidget(groupBox, index // 2, index % 2)

        self.tabCentralWidget.setLayout(self.tabCentralWidgetLayout)

        self.tabWidget = AppQTabWidget(translatable=tabTranslatable)
        self.tabWidget.addTab(self.tabCentralWidget, tabText)

        self.dialogBtns = AppQDialogButtonBox(QtCore.Qt.Orientation.Horizontal)
        self.dialogBtns.addButton(_('OK'), AppQDialogButtonBox.ButtonRole.AcceptRole)
        self.dialogBtns.addButton(
            _('Cancel'), AppQDialogButtonBox.ButtonRole.RejectRole
        )
        self.dialogBtns.accepted.connect(self.accept)
        self.dialogBtns.rejected.connect(self.reject)

        layout = QFormLayout()
        layout.addRow(self.tabWidget)
        layout.addRow(self.dialogBtns)
        layout.setFormAlignment(QtCore.Qt.AlignmentFlag.AlignCenter)

        self.setLayout(layout)

    def groupBoxSequence(self) -> Sequence[GuiEditorWidgetQGroupBox]:
        raise NotImplementedError

    def setTabText(self, text: str):
        self.tabWidget.setTabText(0, text)

    def closeEvent(self, event):
        self.accepted.disconnect()
        self.rejected.disconnect()

    def inputToFactory(self, config: ConfigurationFactory) -> bool:
        modified = False

        for groupBox in self.groupBoxes:
            modified |= groupBox.inputToFactory(config)

        return modified

    def factoryToInput(self, config: ConfigurationFactory):
        for groupBox in self.groupBoxes:
            groupBox.factoryToInput(config)



# Copyright (C) 2024  Loren Eteval <loren.eteval@proton.me>
#
# This file is part of Furious.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

from __future__ import annotations

from Furious.Utility.Utility import BinarySettings

from PySide6 import QtCore

import logging

__all__ = ['AppSettings', 'registerAppSettings']

logger = logging.getLogger(__name__)


class AppSettings:
    SettingsPool: dict[str, AppSettings] = dict()

    def __init__(
        self,
        name: str,
        isBinary=False,
        validRange: list = None,
        default=None,
    ):
        self.name = name
        self.isBinary = isBinary

        if isBinary:
            self.validRange = BinarySettings.RANGE
            self.default = BinarySettings.OFF if default is None else default
        else:
            self.validRange = validRange

            if validRange is None:
                self.default = default
            else:
                self.default = validRange[0] if default is None else default

        if name in AppSettings.SettingsPool:
            raise ValueError(f'\'{name}\' already exists in AppSettings')

    def validate(self, value) -> bool:
        if self.validRange is None:
            return True
        else:
            return value in self.validRange

    @staticmethod
    def get(key: str):
        settings = AppSettings.SettingsPool.get(key)

        if settings is None:
            raise AttributeError(f'AppSettings \'{key}\' not found')

        assert isinstance(settings, AppSettings)

        value = QtCore.QSettings().value(settings.name)

        if settings.validate(value):
            return value
        else:
            logger.error(
                f'settings \'{settings.name}\' has value \'{value}\', '
                f'which is not in valid range {settings.validRange}. '
                f'Set to default \'{settings.default}\''
            )

            # Value not in valid range, set to default
            QtCore.QSettings().setValue(settings.name, settings.default)

            return settings.default

    @staticmethod
    def isStateON_(key: str) -> bool:
        value = AppSettings.get(key)

        if value == BinarySettings.ON_:
            return True
        else:
            return False

    @staticmethod
    def isStateOFF(key: str) -> bool:
        value = AppSettings.get(key)

        if value == BinarySettings.OFF:
            return True
        else:
            return False

    @staticmethod
    def set(key: str, value):
        settings = AppSettings.SettingsPool.get(key)

        if settings is None:
            raise AttributeError(f'AppSettings \'{key}\' not found')

        assert isinstance(settings, AppSettings)

        if settings.validate(value):
            QtCore.QSettings().setValue(settings.name, value)
        else:
            # Value not in valid range, raise exception
            raise ValueError(f'Invalid AppSettings value \'{value}\' for \'{key}\'')

    @staticmethod
    def turnON_(key: str):
        AppSettings.set(key, BinarySettings.ON_)

    @staticmethod
    def turnOFF(key: str):
        AppSettings.set(key, BinarySettings.OFF)


def registerAppSettings(name: str, *args, **kwargs):
    AppSettings.SettingsPool[name] = AppSettings(name, *args, **kwargs)



# Copyright (C) 2024  Loren Eteval <loren.eteval@proton.me>
#
# This file is part of Furious.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

from Furious.PyFramework import *
from Furious.QtFramework import *
from Furious.QtFramework import gettext as _
from Furious.Utility import *

from PySide6 import QtCore
from PySide6.QtWidgets import QHBoxLayout, QProgressBar, QWidget

__all__ = ['ConnectProgressBar']


class ConnectProgressBar(QTranslatable, SupportConnectedCallback, QWidget):
    def __init__(self, parent=None):
        super().__init__(parent)

        self.setWindowTitle(_(APPLICATION_NAME))
        self.setWindowIcon(AppHue.currentWindowIcon())
        self.setFixedSize(280, 61)

        @QtCore.Slot()
        def updateProgressBar():
            # Update the progress bar value
            if self._progressBar.value() < 90:
                self._progressBar.setValue(self._progressBar.value() + 1)

            # Stop the timer when the progress bar reaches 100%
            if self._progressBar.value() >= 100:
                self._timer.stop()

        # Create a progress bar widget
        self._progressBar = QProgressBar(self)
        self._progressBar.setRange(0, 100)
        self._progressBar.setStyleSheet(self.getStyleSheet(AppHue.disconnectedColor()))

        # Create a timer to update the progress bar
        self._timer = QtCore.QTimer(self)
        self._timer.timeout.connect(updateProgressBar)

        self._layout = QHBoxLayout()
        self._layout.addWidget(self._progressBar)

        self.setLayout(self._layout)

    def setValue(self, value: int):
        self._progressBar.setValue(value)

    def start(self, msec: int):
        self._timer.start(msec)

    def stop(self):
        self._timer.stop()

    @staticmethod
    def getStyleSheet(color):
        return (
            f'QProgressBar {{'
            f'    border-radius: 2px;'
            f'    text-align: center;'
            f'}}'
            f''
            f'QProgressBar::chunk {{'
            f'    background-color: {color};'
            f'    width: 10px;'
            f'    margin: 0.5px;'
            f'}}'
        )

    def closeEvent(self, event):
        event.ignore()

        self.hide()

    def disconnectedCallback(self):
        self.setWindowIcon(AppHue.disconnectedWindowIcon())
        self._progressBar.setStyleSheet(self.getStyleSheet(AppHue.disconnectedColor()))

    def connectedCallback(self):
        self.setWindowIcon(AppHue.connectedWindowIcon())
        self._progressBar.setStyleSheet(self.getStyleSheet(AppHue.connectedColor()))

    def retranslate(self):
        self.setWindowTitle(_(self.windowTitle()))



# Copyright (C) 2024  Loren Eteval <loren.eteval@proton.me>
#
# This file is part of Furious.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

from Furious.PyFramework import *

__all__ = ['QBlockSignals', 'QTranslatable']


class QProtection:
    def __init__(self, qobject):
        self.qobject = qobject
        self.hasattr = hasattr(self.qobject, 'setDisabled')

    def __enter__(self):
        if self.hasattr:
            self.qobject.setDisabled(True)

    def __exit__(self, exceptionType, exceptionValue, tb):
        if self.hasattr:
            self.qobject.setDisabled(False)


class QBlockSignals:
    def __init__(self, qobject):
        self.qobject = qobject
        self.hasattr = hasattr(self.qobject, 'blockSignals')

    def __enter__(self):
        if self.hasattr:
            self.qobject.blockSignals(True)

    def __exit__(self, exceptionType, exceptionValue, tb):
        if self.hasattr:
            self.qobject.blockSignals(False)


class QTranslatable(Translatable):
    def __init__(self, *args, **kwargs):
        self.useQProtection = kwargs.pop('useQProtection', True)

        super().__init__(*args, **kwargs)

    def retranslate(self):
        raise NotImplementedError

    @staticmethod
    def retranslateAll():
        for ob in QTranslatable.ObjectsPool:
            assert isinstance(ob, QTranslatable)

            if ob.translatable:
                if ob.useQProtection:
                    with QProtection(ob):
                        ob.retranslate()
                else:
                    ob.retranslate()



# Copyright (C) 2024  Loren Eteval <loren.eteval@proton.me>
#
# This file is part of Furious.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

from __future__ import annotations

from Furious.PyFramework import *
from Furious.QtFramework import *
from Furious.QtFramework import gettext as _
from Furious.Utility import *
from Furious.Library import *
from Furious.Widget.IndentSpinBox import *

from PySide6 import QtCore
from PySide6.QtGui import *
from PySide6.QtWidgets import *

import functools

__all__ = ['TextEditorWindow']

registerAppSettings('ServerWidgetPointSize')

needTrans = functools.partial(needTransFn, source=__name__)

needTrans(
    'Save Changes',
    'The content has been modified. Save changes?',
    'Save',
    'Cancel',
    'Discard',
)


class QuestionSaveMBox(AppQMessageBox):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.setWindowTitle(_('Save Changes'))
        self.setText(_('The content has been modified. Save changes?'))

        self.button0 = self.addButton(_('Save'), AppQMessageBox.ButtonRole.AcceptRole)
        self.button1 = self.addButton(_('Cancel'), AppQMessageBox.ButtonRole.RejectRole)
        self.button2 = self.addButton(
            _('Discard'), AppQMessageBox.ButtonRole.DestructiveRole
        )

        self.setDefaultButton(self.button0)


needTrans('Please check if the configuration is in valid JSON format')


class JSONDecodeErrorMBox(AppQMessageBox):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.error = ''

    def customText(self):
        return (
            _('Please check if the configuration is in valid JSON format')
            + f'\n\n{self.error}'
        )

    def retranslate(self):
        self.setWindowTitle(_(self.windowTitle()))
        self.setText(self.customText())

        # Ignore informative text, buttons

        self.moveToCenter()


needTrans(
    'Save',
    'Save As...',
    'Exit',
    'File',
    'Undo',
    'Redo',
    'Cut',
    'Copy',
    'Paste',
    'Select All',
    'Indent...',
    'Edit',
    'Zoom In',
    'Zoom Out',
    'View',
    'Error saving configuration',
    'Save File',
    'Text files (*.json);;All files (*)',
    'Error Saving File',
    'Invalid server configuration',
    'Error setting indent',
)


class TextEditorWindow(AppQMainWindow):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.customWindowTitle = ''
        self.setWindowModality(QtCore.Qt.WindowModality.ApplicationModal)
        self.setFixedSize(450, int(450 * GOLDEN_RATIO))

        # Current editing index
        self.currentIndex = -1

        self.modified = False
        self.modifiedMark = ' *'

        self.lineColumnLabel = QLabel('1:1 0')
        self.statusBar().addPermanentWidget(self.lineColumnLabel)

        def modificationCallback():
            self.markAsModified()

        def cursorChangedCallback(cursor: QTextCursor):
            self.lineColumnLabel.setText(
                f'{cursor.blockNumber() + 1}:{cursor.columnNumber() + 1} {cursor.position()}'
            )

        self.jsonEditor = DraculaJSONTextEditor(
            fontFamily=APP().customFontName,
            pointSizeSettingsName='ServerWidgetPointSize',
        )
        self.jsonEditor.setLineWrapMode(DraculaJSONTextEditor.LineWrapMode.NoWrap)
        self.jsonEditor.registerModificationChangedCb(modificationCallback)
        self.jsonEditor.registerCursorPositionChangedCb(cursorChangedCallback)

        self.setCentralWidget(self.jsonEditor)

        self.fileMenu = AppQMenu(
            AppQAction(
                _('Save'),
                icon=bootstrapIcon('save.svg'),
                callback=lambda: self.save(),
                shortcut=QtCore.QKeyCombination(
                    QtCore.Qt.KeyboardModifier.ControlModifier,
                    QtCore.Qt.Key.Key_S,
                ),
            ),
            AppQAction(
                _('Save As...'),
                callback=lambda: self.saveAsFile(),
            ),
            AppQSeperator(),
            AppQAction(
                _('Exit'),
                callback=lambda: self.questionSave(),
            ),
            title=_('File'),
            parent=self.menuBar(),
        )

        self.editMenu = AppQMenu(
            AppQAction(
                _('Undo'),
                icon=bootstrapIcon('arrow-return-left.svg'),
                callback=lambda: self.jsonEditor.undo(),
                shortcut=QtCore.QKeyCombination(
                    QtCore.Qt.KeyboardModifier.ControlModifier,
                    QtCore.Qt.Key.Key_Z,
                ),
            ),
            AppQAction(
                _('Redo'),
                icon=bootstrapIcon('arrow-return-right.svg'),
                callback=lambda: self.jsonEditor.redo(),
                shortcut=QtCore.QKeyCombination(
                    QtCore.Qt.KeyboardModifier.ControlModifier
                    | QtCore.Qt.KeyboardModifier.ShiftModifier,
                    QtCore.Qt.Key.Key_Z,
                ),
            ),
            AppQSeperator(),
            AppQAction(
                _('Cut'),
                icon=bootstrapIcon('scissors.svg'),
                callback=lambda: self.jsonEditor.cut(),
                shortcut=QtCore.QKeyCombination(
                    QtCore.Qt.KeyboardModifier.ControlModifier,
                    QtCore.Qt.Key.Key_X,
                ),
            ),
            AppQAction(
                _('Copy'),
                icon=bootstrapIcon('files.svg'),
                callback=lambda: self.jsonEditor.copy(),
                shortcut=QtCore.QKeyCombination(
                    QtCore.Qt.KeyboardModifier.ControlModifier,
                    QtCore.Qt.Key.Key_C,
                ),
            ),
            AppQAction(
                _('Paste'),
                callback=lambda: self.jsonEditor.paste(),
                shortcut=QtCore.QKeyCombination(
                    QtCore.Qt.KeyboardModifier.ControlModifier,
                    QtCore.Qt.Key.Key_V,
                ),
            ),
            AppQSeperator(),
            AppQAction(
                _('Select All'),
                callback=lambda: self.jsonEditor.selectAll(),
                shortcut=QtCore.QKeyCombination(
                    QtCore.Qt.KeyboardModifier.ControlModifier,
                    QtCore.Qt.Key.Key_A,
                ),
            ),
            AppQSeperator(),
            AppQAction(
                _('Indent...'),
                callback=lambda: self.setIndent(),
            ),
            title=_('Edit'),
            parent=self.menuBar(),
        )

        self.viewMenu = AppQMenu(
            AppQAction(
                _('Zoom In'),
                callback=lambda: self.jsonEditor.zoomIn(),
                shortcut=QtCore.QKeyCombination(
                    QtCore.Qt.KeyboardModifier.ControlModifier,
                    QtCore.Qt.Key.Key_Plus,
                ),
            ),
            AppQAction(
                _('Zoom Out'),
                callback=lambda: self.jsonEditor.zoomOut(),
                shortcut=QtCore.QKeyCombination(
                    QtCore.Qt.KeyboardModifier.ControlModifier,
                    QtCore.Qt.Key.Key_Minus,
                ),
            ),
            title=_('View'),
            parent=self.menuBar(),
        )

        self.menuBar().addMenu(self.fileMenu)
        self.menuBar().addMenu(self.editMenu)
        self.menuBar().addMenu(self.viewMenu)

    def markAsModified(self):
        self.modified = True
        self.setWindowTitle(self.customWindowTitle + self.modifiedMark)

    def markAsSaved(self):
        self.modified = False
        self.setWindowTitle(self.customWindowTitle)

    def questionSave(self):
        if self.modified:

            def handleResultCode(code):
                if code == PySide6LegacyEnumValueWrapper(
                    AppQMessageBox.ButtonRole.AcceptRole
                ):
                    if self.save():
                        self.hide()
                if code == PySide6LegacyEnumValueWrapper(
                    AppQMessageBox.ButtonRole.DestructiveRole
                ):
                    self.markAsSaved()
                    self.hide()
                if code == PySide6LegacyEnumValueWrapper(
                    AppQMessageBox.ButtonRole.RejectRole
                ):
                    # Cancel. Do nothing
                    pass

            mbox = QuestionSaveMBox(icon=AppQMessageBox.Icon.Question)
            mbox.finished.connect(functools.partial(handleResultCode))

            # dummy ref
            setattr(self, '_questionSaveMBox', mbox)

            # Show the MessageBox asynchronously
            mbox.open()
        else:
            self.hide()

    def setPlainText(self, text: str, blockSignals: bool):
        if blockSignals:
            with QBlockSignals(self.jsonEditor):
                self.jsonEditor.setPlainText(text)
        else:
            self.jsonEditor.setPlainText(text)

    def save(self) -> bool:
        index = self.currentIndex

        if index < 0:
            # Should not reach here. Do nothing
            self.markAsSaved()

            return True

        plain = self.jsonEditor.toPlainText()

        try:
            jsonObject = JSONEncoder.decode(plain)
        except Exception as ex:
            # Any non-exit exceptions

            mbox = JSONDecodeErrorMBox(icon=AppQMessageBox.Icon.Critical)
            mbox.error = str(ex)
            mbox.setWindowTitle(_('Error saving configuration'))
            mbox.setText(mbox.customText())

            # Show the MessageBox asynchronously
            mbox.open()

            return False
        else:
            old = AS_UserServers()[index]
            new = constructFromDict(jsonObject, **old.kwargs)

            FastItemDeletionSearch.moveToTrash(old)

            AS_UserServers()[index] = new

            try:
                APP().mainWindow.flushRow(index, new)
            except Exception:
                # Any non-exit exceptions

                pass

            if index == AS_UserActivatedItemIndex():
                try:
                    if APP().isSystemTrayConnected():
                        mbox = NewChangesNextTimeMBox()

                        # Show the MessageBox asynchronously
                        mbox.open()
                except Exception:
                    # Any non-exit exceptions

                    pass

            self.markAsSaved()

            return True

    def saveAsFile(self):
        filename, selectedFilter = QFileDialog.getSaveFileName(
            None, _('Save File'), filter=_('Text files (*.json);;All files (*)')
        )

        if filename:
            try:
                with open(filename, 'w', encoding='utf-8') as file:
                    file.write(self.jsonEditor.toPlainText())
            except Exception as ex:
                # Any non-exit exceptions

                mbox = AppQMessageBox(icon=AppQMessageBox.Icon.Critical)
                mbox.setWindowTitle(_('Error Saving File'))
                mbox.setText(_('Invalid server configuration'))
                mbox.setInformativeText(str(ex))

                # Show the MessageBox asynchronously
                mbox.open()

    def setIndent(self):
        def handleResultCode(_indentSpinBox, code):
            if code == PySide6LegacyEnumValueWrapper(AppQDialog.DialogCode.Accepted):
                plain = self.jsonEditor.toPlainText()

                try:
                    jsonObject = JSONEncoder.decode(plain)
                except Exception as ex:
                    # Any non-exit exceptions

                    mbox = JSONDecodeErrorMBox(icon=AppQMessageBox.Icon.Critical)
                    mbox.error = str(ex)
                    mbox.setWindowTitle(_('Error setting indent'))
                    mbox.setText(mbox.customText())

                    # Show the MessageBox asynchronously
                    mbox.open()
                else:
                    text = JSONEncoder.encode(jsonObject, indent=_indentSpinBox.value())

                    self.setPlainText(text, False)
            else:
                # Do nothing
                pass

        indentSpinBox = IndentSpinBox(parent=self)
        indentSpinBox.finished.connect(
            functools.partial(handleResultCode, indentSpinBox)
        )

        # dummy ref
        setattr(self, '_IndentSpinBox', indentSpinBox)

        # Show the MessageBox asynchronously
        indentSpinBox.open()

    def showTabAndSpaces(self):
        textOption = QTextOption()
        textOption.setFlags(QTextOption.Flag.ShowTabsAndSpaces)

        self.jsonEditor.document().setDefaultTextOption(textOption)
        # Reset. Set
        self.jsonEditor.setLineWrapMode(QPlainTextEdit.LineWrapMode.WidgetWidth)
        self.jsonEditor.setLineWrapMode(QPlainTextEdit.LineWrapMode.NoWrap)

    def hideTabAndSpaces(self):
        textOption = QTextOption()

        self.jsonEditor.document().setDefaultTextOption(textOption)
        # Reset. Set
        self.jsonEditor.setLineWrapMode(QPlainTextEdit.LineWrapMode.WidgetWidth)
        self.jsonEditor.setLineWrapMode(QPlainTextEdit.LineWrapMode.NoWrap)

    def closeEvent(self, event: QtCore.QEvent):
        event.ignore()

        self.questionSave()

    def retranslate(self):
        # Do nothing
        pass



# Copyright (C) 2024  Loren Eteval <loren.eteval@proton.me>
#
# This file is part of Furious.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

from __future__ import annotations

from Furious.Interface import *
from Furious.QtFramework import *
from Furious.QtFramework import gettext as _
from Furious.Utility import *

from PySide6.QtWidgets import *

from typing import Callable

__all__ = ['GuiVTLSQGroupBox']

needTrans = functools.partial(needTransFn, source=__name__)

STREAM_SECURITY = [
    '',
    'none',
    'tls',
    'reality',
]


class GuiVTLSItemSecurity(GuiEditorItemTextComboBox):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.addItems(STREAM_SECURITY)

    def inputToFactory(self, config: ConfigurationFactory) -> bool:
        streamSettings = getXrayProxyOutboundStream(config)

        try:
            oldSecurity = streamSettings['security']
        except Exception:
            # Any non-exit exceptions

            oldSecurity = ''

        newSecurity = self.text()

        def setNewSecurity():
            if newSecurity == '':
                streamSettings.pop('security', None)
            else:
                streamSettings['security'] = newSecurity

            for security in ['tls', 'reality']:
                if security == newSecurity:
                    continue

                securityKey = f'{security}Settings'

                # Remove irrelevant settings
                streamSettings.pop(securityKey, None)

        if isinstance(oldSecurity, str):
            if newSecurity != oldSecurity:
                setNewSecurity()

                return True
            else:
                return False
        else:
            setNewSecurity()

            return True

    def factoryToInput(self, config: ConfigurationFactory):
        try:
            streamSettings = getXrayProxyOutboundStream(config)

            self.setText(streamSettings.get('security', ''))
        except Exception:
            # Any non-exit exceptions

            self.setText('')


class GuiVTLSItemXXXServerName(GuiEditorItemTextInput):
    def __init__(self, *args, **kwargs):
        # Mandatory
        securityKey = kwargs.pop('securityKey')

        super().__init__(*args, **kwargs)

        self.securityKey = securityKey

    def inputToFactory(self, config: ConfigurationFactory) -> bool:
        streamSettings = getXrayProxyOutboundStream(config)

        if not isinstance(streamSettings.get(self.securityKey), dict):
            streamSettings[self.securityKey] = {}

        xxxObject = streamSettings[self.securityKey]

        try:
            oldServerName = xxxObject.get('serverName', '')
        except Exception:
            # Any non-exit exceptions

            oldServerName = ''

        newServerName = self.text()

        def setNewServerName():
            if newServerName == '':
                xxxObject.pop('serverName', None)
            else:
                xxxObject['serverName'] = newServerName

        if isinstance(oldServerName, str):
            if newServerName != oldServerName:
                setNewServerName()

                return True
            else:
                return False
        else:
            setNewServerName()

            return True

    def factoryToInput(self, config: ConfigurationFactory):
        try:
            xxxObject = getXrayProxyOutboundStream(config)[self.securityKey]

            self.setText(xxxObject.get('serverName', ''))
        except Exception:
            # Any non-exit exceptions

            self.setText('')


class GuiVTLSItemTLSServerName(GuiVTLSItemXXXServerName):
    def __init__(self, *args, **kwargs):
        securityKey = kwargs.pop('securityKey', 'tlsSettings')

        super().__init__(*args, **kwargs, securityKey=securityKey)


class GuiVTLSItemRealityServerName(GuiVTLSItemXXXServerName):
    def __init__(self, *args, **kwargs):
        securityKey = kwargs.pop('securityKey', 'realitySettings')

        super().__init__(*args, **kwargs, securityKey=securityKey)


class GuiVTLSItemXXXFingerprint(GuiEditorItemTextInput):
    def __init__(self, *args, **kwargs):
        # Mandatory
        securityKey = kwargs.pop('securityKey')

        super().__init__(*args, **kwargs)

        self.securityKey = securityKey

    def inputToFactory(self, config: ConfigurationFactory) -> bool:
        streamSettings = getXrayProxyOutboundStream(config)

        if not isinstance(streamSettings.get(self.securityKey), dict):
            streamSettings[self.securityKey] = {}

        xxxObject = streamSettings[self.securityKey]

        try:
            oldFingerprint = xxxObject.get('fingerprint', '')
        except Exception:
            # Any non-exit exceptions

            oldFingerprint = ''

        newFingerprint = self.text()

        def setNewFingerprint():
            if newFingerprint == '':
                xxxObject.pop('fingerprint', None)
            else:
                xxxObject['fingerprint'] = newFingerprint

        if isinstance(oldFingerprint, str):
            if newFingerprint != oldFingerprint:
                setNewFingerprint()

                return True
            else:
                return False
        else:
            setNewFingerprint()

            return True

    def factoryToInput(self, config: ConfigurationFactory):
        try:
            xxxObject = getXrayProxyOutboundStream(config)[self.securityKey]

            self.setText(xxxObject.get('fingerprint', ''))
        except Exception:
            # Any non-exit exceptions

            self.setText('')


class GuiVTLSItemTLSFingerprint(GuiVTLSItemXXXFingerprint):
    def __init__(self, *args, **kwargs):
        securityKey = kwargs.pop('securityKey', 'tlsSettings')

        super().__init__(*args, **kwargs, securityKey=securityKey)


class GuiVTLSItemRealityFingerprint(GuiVTLSItemXXXFingerprint):
    def __init__(self, *args, **kwargs):
        securityKey = kwargs.pop('securityKey', 'realitySettings')

        super().__init__(*args, **kwargs, securityKey=securityKey)


class GuiVTLSItemTLSAlpn(GuiEditorItemTextInput):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def inputToFactory(self, config: ConfigurationFactory) -> bool:
        streamSettings = getXrayProxyOutboundStream(config)

        if not isinstance(streamSettings.get('tlsSettings'), dict):
            streamSettings['tlsSettings'] = {}

        tlsObject = streamSettings['tlsSettings']

        try:
            oldAlpn = ','.join(tlsObject.get('alpn', []))
        except Exception:
            # Any non-exit exceptions

            oldAlpn = ''

        newAlpn = self.text()

        def setNewAlpn():
            if newAlpn == '':
                tlsObject.pop('alpn', None)
            else:
                tlsObject['alpn'] = newAlpn.split(',')

        if isinstance(oldAlpn, str):
            if newAlpn != oldAlpn:
                setNewAlpn()

                return True
            else:
                return False
        else:
            setNewAlpn()

            return True

    def factoryToInput(self, config: ConfigurationFactory):
        try:
            tlsObject = getXrayProxyOutboundStream(config)['tlsSettings']

            self.setText(','.join(tlsObject.get('alpn', [])))
        except Exception:
            # Any non-exit exceptions

            self.setText('')


class GuiVTLSItemTLSAllowInsecure(GuiEditorItemTextCheckBox):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def inputToFactory(self, config: ConfigurationFactory) -> bool:
        streamSettings = getXrayProxyOutboundStream(config)

        if not isinstance(streamSettings.get('tlsSettings'), dict):
            streamSettings['tlsSettings'] = {}

        tlsObject = streamSettings['tlsSettings']

        try:
            oldInsecure = tlsObject.get('allowInsecure', False)
        except Exception:
            # Any non-exit exceptions

            oldInsecure = False

        newInsecure = self.isChecked()

        if newInsecure:
            if oldInsecure is not True:
                tlsObject['allowInsecure'] = True

                return True
            else:
                return False
        else:
            if oldInsecure is not False:
                tlsObject['allowInsecure'] = False

                return True
            else:
                return False

    def factoryToInput(self, config: ConfigurationFactory):
        try:
            tlsObject = getXrayProxyOutboundStream(config)['tlsSettings']

            self.setChecked(tlsObject.get('allowInsecure', False))
        except Exception:
            # Any non-exit exceptions

            self.setChecked(False)


class GuiVTLSItemRealityXXX(GuiEditorItemTextInput):
    def __init__(self, *args, **kwargs):
        # Mandatory
        realityKey = kwargs.pop('realityKey')

        super().__init__(*args, **kwargs)

        self.realityKey = realityKey

    def inputToFactory(self, config: ConfigurationFactory) -> bool:
        streamSettings = getXrayProxyOutboundStream(config)

        if not isinstance(streamSettings.get('realitySettings'), dict):
            streamSettings['realitySettings'] = {}

        realityObject = streamSettings['realitySettings']

        try:
            oldPublicKey = realityObject.get(self.realityKey, '')
        except Exception:
            # Any non-exit exceptions

            oldPublicKey = ''

        newPublicKey = self.text()

        def setNewPublicKey():
            if newPublicKey == '':
                realityObject.pop(self.realityKey, None)
            else:
                realityObject[self.realityKey] = newPublicKey

        if isinstance(oldPublicKey, str):
            if newPublicKey != oldPublicKey:
                setNewPublicKey()

                return True
            else:
                return False
        else:
            setNewPublicKey()

            return True

    def factoryToInput(self, config: ConfigurationFactory):
        try:
            realityObject = getXrayProxyOutboundStream(config)['realitySettings']

            self.setText(realityObject.get(self.realityKey, ''))
        except Exception:
            # Any non-exit exceptions

            self.setText('')


class GuiVTLSItemRealityPublicKey(GuiVTLSItemRealityXXX):
    def __init__(self, *args, **kwargs):
        realityKey = kwargs.pop('realityKey', 'publicKey')

        super().__init__(*args, **kwargs, realityKey=realityKey)


class GuiVTLSItemRealityShortId(GuiVTLSItemRealityXXX):
    def __init__(self, *args, **kwargs):
        realityKey = kwargs.pop('realityKey', 'shortId')

        super().__init__(*args, **kwargs, realityKey=realityKey)


class GuiVTLSItemRealitySpiderX(GuiVTLSItemRealityXXX):
    def __init__(self, *args, **kwargs):
        realityKey = kwargs.pop('realityKey', 'spiderX')

        super().__init__(*args, **kwargs, realityKey=realityKey)


class GuiVTLSPageXXX(GuiEditorWidgetQWidget):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def setSecurityText(self, text: str):
        security = self._containers[0]

        if isinstance(security, GuiEditorItemTextComboBox):
            security.setText(text)

    def connectActivated(self, func: Callable):
        security = self._containers[0]

        if isinstance(security, GuiEditorItemTextComboBox):
            security.connectActivated(func)


class GuiVTLSPageEmpty(GuiVTLSPageXXX):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def containerSequence(self):
        return [
            GuiVTLSItemSecurity(title='TLS', translatable=False),
        ]


class GuiVTLSPageNone(GuiVTLSPageXXX):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def containerSequence(self):
        return [
            GuiVTLSItemSecurity(title='TLS', translatable=False),
        ]


class GuiVTLSPageTLS(GuiVTLSPageXXX):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def containerSequence(self):
        return [
            GuiVTLSItemSecurity(title='TLS', translatable=False),
            GuiVTLSItemTLSServerName(title='SNI', translatable=False),
            GuiVTLSItemTLSFingerprint(title='Fingerprint', translatable=False),
            GuiVTLSItemTLSAlpn(title='Alpn', translatable=False),
            GuiVTLSItemTLSAllowInsecure(title='AllowInsecure', translatable=False),
        ]


class GuiVTLSPageReality(GuiVTLSPageXXX):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def containerSequence(self):
        return [
            GuiVTLSItemSecurity(title='TLS', translatable=False),
            GuiVTLSItemRealityServerName(title='SNI', translatable=False),
            GuiVTLSItemRealityFingerprint(title='Fingerprint', translatable=False),
            GuiVTLSItemRealityPublicKey(title='PublicKey', translatable=False),
            GuiVTLSItemRealityShortId(title='ShortId', translatable=False),
            GuiVTLSItemRealitySpiderX(title='SpiderX', translatable=False),
        ]


class GuiVTLSPageStackedWidget(QStackedWidget):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        # Corresponds to stream security
        self._pages = [
            GuiVTLSPageEmpty(),
            GuiVTLSPageNone(),
            GuiVTLSPageTLS(),
            GuiVTLSPageReality(),
        ]

        for page in self._pages:
            self.addWidget(page)

    def page(self, index: int) -> GuiVTLSPageXXX:
        return self._pages[index]

    def connectActivated(self, func: Callable):
        for page in self._pages:
            page.connectActivated(func)


class GuiVTLSQGroupBox(GuiEditorItemFactory, AppQGroupBox):
    def __init__(self, **kwargs):
        translatable = kwargs.pop('translatable', False)

        super().__init__('TLS', **kwargs, translatable=translatable)

        self._config = ConfigurationFactory()

        self._widget = GuiVTLSPageStackedWidget()
        self._widget.connectActivated(self.handleActivated)

        layout = QFormLayout()
        layout.addRow(self._widget)

        self.setLayout(layout)

    def currentIndex(self) -> int:
        return self._widget.currentIndex()

    def setCurrentIndex(self, index: int):
        self._widget.setCurrentIndex(index)

    def page(self, index: int) -> GuiVTLSPageXXX:
        return self._widget.page(index)

    def handleActivated(self, index: int):
        page = self.page(index)
        page.factoryToInput(self._config)
        page.setSecurityText(STREAM_SECURITY[index])

        self.setCurrentIndex(index)

    def inputToFactory(self, config: ConfigurationFactory) -> bool:
        return self.page(self.currentIndex()).inputToFactory(config)

    def factoryToInput(self, config: ConfigurationFactory):
        # Shallow copy
        self._config = config

        streamSettings = getXrayProxyOutboundStream(config)
        security = streamSettings.get('security', '')

        if not isinstance(security, str):
            return

        try:
            index = STREAM_SECURITY.index(security)
        except Exception:
            # Any non-exit exceptions

            pass
        else:
            self.page(index).factoryToInput(config)
            self.setCurrentIndex(index)



# Copyright (C) 2024  Loren Eteval <loren.eteval@proton.me>
#
# This file is part of Furious.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

from __future__ import annotations

from Furious.PyFramework import *
from Furious.QtFramework import *
from Furious.QtFramework import gettext as _
from Furious.Utility import *

from PySide6 import QtCore
from PySide6.QtWidgets import *

import os
import shutil
import logging
import functools
import darkdetect

__all__ = ['XrayAssetViewerQListWidget']

logger = logging.getLogger(__name__)

needTrans = functools.partial(needTransFn, source=__name__)


class AssetExistsMBox(AppQMessageBox):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.setStandardButtons(
            AppQMessageBox.StandardButton.Yes | AppQMessageBox.StandardButton.No
        )

    def retranslate(self):
        self.setWindowTitle(_(self.windowTitle()))
        self.setText(_(self.text()))

        # Ignore informative text, buttons

        self.moveToCenter()


needTrans(
    'Delete',
    'Import',
    'Asset file already exists. Overwrite?',
    'Error import asset file',
    'Import asset file success',
)


class XrayAssetViewerQListWidget(SupportThemeChangedCallback, AppQListWidget):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.setSelectionBehavior(AppQListWidget.SelectionBehavior.SelectRows)
        self.setSelectionMode(AppQListWidget.SelectionMode.ExtendedSelection)
        self.setIconSize(QtCore.QSize(64, 64))

        if PLATFORM == 'Linux' and getUbuntuRelease() == '20.04':
            self.initialTheme = darkdetect.theme()
        else:
            self.initialTheme = None

        self.flushItem()

        logger.info(f'Xray-core asset dir is \'{XRAY_ASSET_DIR}\'')

        contextMenuActions = [
            AppQAction(
                _('Delete'),
                callback=lambda: self.deleteSelectedItem(),
            ),
        ]

        self.contextMenu = AppQMenu(*contextMenuActions)

        # Add actions to self in order to activate shortcuts
        self.addActions(self.contextMenu.actions())

        self.setContextMenuPolicy(QtCore.Qt.ContextMenuPolicy.CustomContextMenu)
        self.customContextMenuRequested.connect(self.handleCustomContextMenuRequested)

    @QtCore.Slot(QtCore.QPoint)
    def handleCustomContextMenuRequested(self, point):
        self.contextMenu.exec(self.mapToGlobal(point))

    def flushItemByTheme(self, theme: str):
        self.clear()

        for filename in os.listdir(XRAY_ASSET_DIR):
            if os.path.isfile(XRAY_ASSET_DIR / filename):
                item = QListWidgetItem(filename)

                if AppSettings.isStateON_('DarkMode'):
                    # Custom dark mode
                    item.setIcon(bootstrapIconWhite('file-earmark.svg'))
                else:
                    if theme == 'Dark':
                        if PLATFORM == 'Windows':
                            # Windows. Always use black icon
                            item.setIcon(bootstrapIcon('file-earmark.svg'))
                        else:
                            item.setIcon(bootstrapIconWhite('file-earmark.svg'))
                    else:
                        item.setIcon(bootstrapIcon('file-earmark.svg'))

                self.addItem(item)

    def flushItem(self):
        if PLATFORM == 'Linux' and getUbuntuRelease() == '20.04':
            assert self.initialTheme is not None

            # Ubuntu 20.04. Flush by initial theme(Ubuntu 20.04 theme changes bug)
            self.flushItemByTheme(self.initialTheme)
        else:
            self.flushItemByTheme(darkdetect.theme())

    def appendNewItem(self, filename: str):
        def append(_filename):
            try:
                shutil.copy(_filename, XRAY_ASSET_DIR)
            except shutil.SameFileError:
                # Same file imported. Do nothing
                pass
            except Exception as ex:
                # Any non-exit exception

                _mbox = AppQMessageBox(icon=AppQMessageBox.Icon.Critical)
                _mbox.setWindowTitle(_('Import'))
                _mbox.setText(_('Error import asset file'))
                _mbox.setInformativeText(str(ex))

                # Show the MessageBox asynchronously
                _mbox.open()
            else:
                self.flushItem()

                _mbox = AppQMessageBox(icon=AppQMessageBox.Icon.Information)
                _mbox.setWindowTitle(_('Import'))
                _mbox.setText(_('Import asset file success'))

                # Show the MessageBox asynchronously
                _mbox.open()

        basename = os.path.basename(filename)

        if os.path.isfile(XRAY_ASSET_DIR / basename):

            def handleResultCode(_filename, code):
                if code == PySide6LegacyEnumValueWrapper(
                    AppQMessageBox.StandardButton.Yes
                ):
                    append(_filename)
                else:
                    # Do not overwrite
                    pass

            mbox = AssetExistsMBox(icon=AppQMessageBox.Icon.Question)
            mbox.setWindowTitle(_('Import'))
            mbox.setText(_('Asset file already exists. Overwrite?'))
            mbox.setInformativeText(basename)
            mbox.finished.connect(functools.partial(handleResultCode, filename))

            # dummy ref
            setattr(self, '_assetExistsMBox', mbox)

            # Show the MessageBox asynchronously
            mbox.open()
        else:
            append(filename)

    def deleteSelectedItem(self):
        indexes = self.selectedIndex

        if len(indexes) == 0:
            # Nothing selected
            return

        def handleResultCode(_indexes, code):
            if code == PySide6LegacyEnumValueWrapper(AppQMessageBox.StandardButton.Yes):
                for index in _indexes:
                    os.remove(XRAY_ASSET_DIR / self.item(index).text())

                self.flushItem()
            else:
                # Do not delete
                pass

        mbox = QuestionDeleteMBox(icon=AppQMessageBox.Icon.Question)
        mbox.isMulti = bool(len(indexes) > 1)
        mbox.possibleRemark = f'{self.item(indexes[0]).text()}'
        mbox.setText(mbox.customText())
        mbox.finished.connect(functools.partial(handleResultCode, indexes))

        # dummy ref
        setattr(self, '_questionDeleteMBox', mbox)

        # Show the MessageBox asynchronously
        mbox.open()

    def keyPressEvent(self, event):
        if event.key() == QtCore.Qt.Key.Key_Delete:
            self.deleteSelectedItem()
        else:
            super().keyPressEvent(event)

    def themeChangedCallback(self, theme):
        if PLATFORM == 'Linux' and getUbuntuRelease() == '20.04':
            # Ubuntu 20.04 system dark theme does not
            # change menu color. Do nothing
            pass
        else:
            self.flushItemByTheme(theme)



# Copyright (C) 2024  Loren Eteval <loren.eteval@proton.me>
#
# This file is part of Furious.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

from Furious.Utility.Constants import *
from Furious.Utility.Utility import *

import time

__all__ = ['PySide6LegacyEnumValueWrapper', 'PySide6LegacyEventLoopWait']


def PySide6LegacyEnumValueWrapper(enum):
    # Protect PySide6 enum wrapper behavior changes
    if versionToValue(PYSIDE6_VERSION) < versionToValue('6.2.2'):
        return enum
    else:
        return enum.value


def PySide6LegacyEventLoopWait(ms):
    # Protect qWait method does not exist in some
    # old PySide6 version
    if versionToValue(PYSIDE6_VERSION) < versionToValue('6.3.1'):
        for counter in range(0, ms, 10):
            time.sleep(10 / 1000)

            APP().processEvents()
    else:
        from PySide6.QtTest import QTest

        QTest.qWait(ms)

        APP().processEvents()



# Copyright (C) 2024  Loren Eteval <loren.eteval@proton.me>
#
# This file is part of Furious.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

from Furious.Interface import *
from Furious.QtFramework import *
from Furious.QtFramework import gettext as _
from Furious.Library import *
from Furious.Utility import *

from PySide6.QtWidgets import QApplication, QFileDialog

import os
import logging
import functools

__all__ = [
    'ImportFromFileAction',
    'ImportURIFromClipboardAction',
    'ImportJSONFromClipboardAction',
    'ImportAction',
]

logger = logging.getLogger(__name__)

needTrans = functools.partial(needTransFn, source=__name__)

needTrans(
    'Invalid data',
    'Invalid data. The content of the clipboard is:',
)


def showImportErrorMBox(clipboard: str):
    mbox = ImportErrorMBox(icon=AppQMessageBox.Icon.Critical)

    if len(clipboard) > 1000:
        # Limited
        mbox.setText(_('Invalid data'))
        mbox.setInformativeText('')
    else:
        mbox.setText(_('Invalid data. The content of the clipboard is:'))
        mbox.setInformativeText(clipboard)

    # Show the MessageBox asynchronously
    mbox.open()


def importItemFromClipboard(clipboard: str):
    factory = constructFromAny(clipboard)

    if not factory.isValid():
        showImportErrorMBox(clipboard)
    else:
        APP().mainWindow.appendNewItemByFactory(factory)

        mbox = ImportSuccessMBox(icon=AppQMessageBox.Icon.Information)
        mbox.remark = factory.getExtras('remark')
        mbox.setText(mbox.customText())

        # Show the MessageBox asynchronously
        mbox.open()


needTrans('Import')


class ImportErrorMBox(AppQMessageBox):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.setWindowTitle(_('Import'))

    def retranslate(self):
        self.setWindowTitle(_(self.windowTitle()))
        self.setText(_(self.text()))

        # Ignore informative text, buttons

        self.moveToCenter()


needTrans(
    'Import',
    'Import share link success',
    'Imported to row',
)


class ImportMultiSuccessMBox(AppQMessageBox):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.imported = []
        self.rowCount = 0

        self.setWindowTitle(_('Import'))
        self.setIcon(AppQMessageBox.Icon.Information)

    def customText(self):
        text = (
            _('Import share link success')
            + f'\n\n'
            + '\n'.join(
                list(
                    f'{index + 1} - {remark}. '
                    + _('Imported to row')
                    + f' {self.rowCount + index + 1}'
                    for index, remark in enumerate(self.imported)
                )
            )
        )

        if len(text) <= 1000:
            return text
        else:
            # Limited
            return _('Import share link success') + f'\n\n...'

    def retranslate(self):
        self.setWindowTitle(_(self.windowTitle()))
        self.setText(self.customText())

        # Ignore informative text, buttons

        self.moveToCenter()


needTrans(
    'Import',
    'Import success',
)


class ImportSuccessMBox(AppQMessageBox):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.remark = ''
        self.setWindowTitle(_('Import'))

    def customText(self):
        if self.remark:
            return _('Import success') + f': {self.remark}'
        else:
            return _('Import success')

    def retranslate(self):
        self.setWindowTitle(_(self.windowTitle()))
        self.setText(self.customText())

        # Ignore informative text, buttons

        self.moveToCenter()


needTrans(
    'Import From File...',
    'Import File',
    'Text files (*.json);;All files (*)',
    'Error opening file',
    'Invalid configuration file',
    'Invalid data',
)


class ImportFromFileAction(AppQAction):
    def __init__(self, **kwargs):
        super().__init__(
            _('Import From File...'),
            icon=bootstrapIcon('folder2-open.svg'),
            **kwargs,
        )

    def triggeredCallback(self, checked):
        filename, selectedFilter = QFileDialog.getOpenFileName(
            None,
            _('Import File'),
            filter=_('Text files (*.json);;All files (*)'),
        )

        if filename:
            try:
                with open(filename, 'r', encoding='utf-8') as file:
                    plainText = file.read()
            except Exception as ex:
                # Any non-exit exceptions

                mbox = AppQMessageBox(icon=AppQMessageBox.Icon.Critical)
                mbox.setWindowTitle(_('Error opening file'))
                mbox.setText(_('Invalid configuration file'))
                mbox.setInformativeText(str(ex))

                # Show the MessageBox asynchronously
                mbox.open()
            else:
                factory = constructFromAny(plainText, remark=os.path.basename(filename))

                if factory.isValid():
                    APP().mainWindow.appendNewItemByFactory(factory)

                    mbox = ImportSuccessMBox(icon=AppQMessageBox.Icon.Information)
                    mbox.remark = factory.getExtras('remark')
                    mbox.setText(mbox.customText())

                    # Show the MessageBox asynchronously
                    mbox.open()
                else:
                    mbox = ImportErrorMBox(icon=AppQMessageBox.Icon.Critical)
                    mbox.setText(_('Invalid data'))
                    mbox.setInformativeText('')

                    # Show the MessageBox asynchronously
                    mbox.open()


needTrans('Import Share Link From Clipboard')


class ImportURIFromClipboardAction(AppQAction):
    def __init__(self, **kwargs):
        super().__init__(_('Import Share Link From Clipboard'), **kwargs)

    def triggeredCallback(self, checked):
        clipboard = QApplication.clipboard().text().strip()

        try:
            split = clipboard.split('\n')
        except Exception:
            # Any non-exit exceptions

            importItemFromClipboard(clipboard)
        else:
            imported = list()
            rowCount = len(AS_UserServers())

            for uri in split:
                factory = constructFromAny(uri)

                if factory.isValid():
                    APP().mainWindow.appendNewItemByFactory(factory)

                    imported.append(factory.getExtras('remark'))

            if len(imported) == 0:
                showImportErrorMBox(clipboard)
            else:
                if len(imported) == 1:
                    # Fall back to single
                    mbox = ImportSuccessMBox(icon=AppQMessageBox.Icon.Information)
                    mbox.remark = imported[0]
                    mbox.setText(mbox.customText())

                    # Show the MessageBox asynchronously
                    mbox.open()
                else:
                    mbox = ImportMultiSuccessMBox(icon=AppQMessageBox.Icon.Information)
                    mbox.imported = imported
                    mbox.rowCount = rowCount
                    mbox.setText(mbox.customText())

                    # Show the MessageBox asynchronously
                    mbox.open()


needTrans('Import JSON Configuration From Clipboard')


class ImportJSONFromClipboardAction(AppQAction):
    def __init__(self, **kwargs):
        super().__init__(_('Import JSON Configuration From Clipboard'), **kwargs)

    def triggeredCallback(self, checked):
        clipboard = QApplication.clipboard().text().strip()

        importItemFromClipboard(clipboard)


needTrans('Import')


class ImportAction(AppQAction):
    def __init__(self, **kwargs):
        super().__init__(
            _('Import'),
            icon=bootstrapIcon('lightning-charge.svg'),
            menu=AppQMenu(
                ImportURIFromClipboardAction(),
                ImportJSONFromClipboardAction(),
            ),
            useActionGroup=False,
            checkable=True,
            **kwargs,
        )



# Copyright (C) 2024  Loren Eteval <loren.eteval@proton.me>
#
# This file is part of Furious.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

from __future__ import annotations

from Furious.Utility import *
from Furious.Library import *
from Furious.Externals import *

import functools

__all__ = [
    'gettext',
    'needTransFn',
    'LANGUAGE_TO_ABBR',
    'ABBR_TO_LANGUAGE',
    'SUPPORTED_LANGUAGE',
    'TranslationPool',
]


class Translator:
    def __init__(self):
        super().__init__()

        self.translation = dict()
        self.dictEnglish = dict()

    def install(self, translation):
        self.translation = translation

        for key, value in translation.items():
            # English -> English
            value['EN'] = key

            for lang, text in value.items():
                if lang.isupper():
                    # is 'ZH', 'EN', etc.
                    self.dictEnglish[text] = key

        # English -> English
        self.dictEnglish.update(dict(list((key, key) for key in translation.keys())))

    def translate(self, source, locale):
        try:
            return self.translation[self.dictEnglish[source]][locale]
        except Exception:
            # Any non-exit exceptions

            # No changes
            return source


translator = Translator()


def installTranslation(translation):
    translator.install(translation)


def gettext(source, locale=None):
    if locale is None:
        assert APP() is not None

        return translator.translate(source, AppSettings.get('Language'))
    else:
        assert locale in SUPPORTED_LANGUAGE

        return translator.translate(source, locale)


# Register new translation type here
LANGUAGE_TO_ABBR = {
    'English': 'EN',
    'Российский': 'RU',
    '简体中文': 'ZH',
}

ABBR_TO_LANGUAGE = {value: key for key, value in LANGUAGE_TO_ABBR.items()}

SUPPORTED_LANGUAGE = list(LANGUAGE_TO_ABBR.values())

installTranslation(TRANSLATION)


class TranslatorHelper:
    TranslationPool = list()

    @staticmethod
    def appendText(*texts, **kwargs):
        source = kwargs.pop('source', '')

        for text in texts:
            if isinstance(text, str):
                TranslatorHelper.TranslationPool.append([text, source])


needTransFn = functools.partial(TranslatorHelper.appendText)
TranslationPool = TranslatorHelper.TranslationPool



# Copyright (C) 2024  Loren Eteval <loren.eteval@proton.me>
#
# This file is part of Furious.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

from __future__ import annotations

from Furious.Utility.Constants import *

from typing import AnyStr, Tuple

import os
import ujson
import operator
import functools
import ipaddress
import subprocess
import urllib.parse

__all__ = [
    'Protocol',
    'BinarySettings',
    'protocolRepr',
    'isValidIPAddress',
    'parseHostPort',
    'runExternalCommand',
    'getAbsolutePath',
    'versionToValue',
    'getXrayProxyOutboundObject',
    'getXrayProxyOutboundStream',
]


class Protocol:
    VMess = 'VMess'
    VLESS = 'VLESS'
    Shadowsocks = 'Shadowsocks'
    Trojan = 'Trojan'
    Hysteria1 = 'hysteria1'
    Hysteria2 = 'hysteria2'


class BinarySettings:
    OFF = '0'
    ON_ = '1'

    RANGE = [OFF, ON_]


@functools.lru_cache(None)
def protocolRepr(protocol: str) -> str:
    if not isinstance(protocol, str):
        return ''

    if protocol.lower() == 'vmess':
        return Protocol.VMess

    if protocol.lower() == 'vless':
        return Protocol.VLESS

    if protocol.lower() == 'shadowsocks':
        return Protocol.Shadowsocks

    if protocol.lower() == 'trojan':
        return Protocol.Trojan

    return protocol


@functools.lru_cache(None)
def isValidIPAddress(address) -> bool:
    try:
        ipaddress.ip_address(address)
    except Exception:
        # Any non-exita exceptions

        return False
    else:
        return True


# Can throw exceptions
def parseHostPort(address) -> Tuple[AnyStr | None, str]:
    if address.count('//') == 0:
        result = urllib.parse.urlsplit('//' + address)
    else:
        result = urllib.parse.urlsplit(address)

    return result.hostname, str(result.port)


def runExternalCommand(*args, **kwargs):
    if PLATFORM == 'Windows':
        creationflags = kwargs.pop('creationflags', subprocess.CREATE_NO_WINDOW)

        return subprocess.run(*args, creationflags=creationflags, **kwargs)
    else:
        return subprocess.run(*args, **kwargs)


def getAbsolutePath(path):
    return path if os.path.isabs(path) else str(ROOT_DIR / path)


def versionToValue(version: str) -> int:
    def _split():
        # x.y.z or x.y.z.u
        result = version.split('.')

        if len(result) == 3:
            result.append('0')

        return result

    x_weight = 10**9
    y_weight = 10**6
    z_weight = 10**3
    u_weight = 1

    try:
        x, y, z, u = _split()

        return functools.reduce(
            operator.add,
            list(
                int(val) * weight
                for val, weight in zip(
                    [x, y, z, u], [x_weight, y_weight, z_weight, u_weight]
                )
            ),
        )
    except Exception:
        # Any non-exit exceptions

        return 0


def getXrayProxyOutboundObject(config: dict) -> dict:
    if not isinstance(config.get('outbounds'), list):
        config['outbounds'] = []

    for outbound in config['outbounds']:
        if isinstance(outbound, dict):
            tag = outbound.get('tag')

            if isinstance(tag, str) and tag == 'proxy':
                return outbound

    return {}


def getXrayProxyOutboundStream(config: dict) -> dict:
    proxyOutboundObject = getXrayProxyOutboundObject(config)

    if not isinstance(proxyOutboundObject.get('streamSettings'), dict):
        proxyOutboundObject['streamSettings'] = {}

    return proxyOutboundObject['streamSettings']



# Copyright (C) 2024  Loren Eteval <loren.eteval@proton.me>
#
# This file is part of Furious.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

from Furious.PyFramework import *
from Furious.QtFramework.QtNetwork import *
from Furious.Utility import *

from PySide6 import QtCore
from PySide6.QtNetwork import *

import logging
import functools

__all__ = ['NetworkStateManager']

logger = logging.getLogger(__name__)


class NetworkStateManager(SupportConnectedCallback, AppQNetworkAccessManager):
    MIN_JOB_INTERVAL = 2500
    MAX_JOB_INTERVAL = 2000000000

    def __init__(self, parent=None):
        super().__init__(parent)

        self.jobStatus = False
        self.jobInterval = NetworkStateManager.MIN_JOB_INTERVAL

        self.jobTimeoutTimer = QtCore.QTimer()
        self.jobArrangeTimer = QtCore.QTimer()

        self.jobArrangeTimer.timeout.connect(lambda: self.startSingleTest())

    def successCallback(self):
        raise NotImplementedError

    def errorCallback(self, errorString: str):
        raise NotImplementedError

    def handleFinishedByNetworkReply(self, networkReply):
        assert isinstance(networkReply, QNetworkReply)

        if networkReply.error() != QNetworkReply.NetworkError.NoError:
            self.jobTimeoutTimer.stop()

            errorString = networkReply.errorString()

            logger.error(f'connection test failed. {errorString}')

            self.errorCallback(errorString)

            if self.jobStatus is False:
                self.jobInterval *= 2
            else:
                self.jobInterval = NetworkStateManager.MIN_JOB_INTERVAL

            self.jobStatus = False
        else:
            self.jobTimeoutTimer.stop()

            logger.info(f'connection test success')

            self.successCallback()

            if self.jobStatus is True:
                self.jobInterval *= 2
            else:
                self.jobInterval = NetworkStateManager.MIN_JOB_INTERVAL

            self.jobStatus = True

        if self.jobInterval >= NetworkStateManager.MAX_JOB_INTERVAL:
            # Limited
            self.jobInterval = NetworkStateManager.MAX_JOB_INTERVAL

        self.jobArrangeTimer.start(self.jobInterval)

    def startSingleTest(self):
        networkReply = self.get(QNetworkRequest(QtCore.QUrl(NETWORK_STATE_TEST_URL)))
        networkReply.finished.connect(
            functools.partial(
                self.handleFinishedByNetworkReply,
                networkReply,
            )
        )

        def abort(_networkReply):
            if isinstance(_networkReply, QNetworkReply):
                _networkReply.abort()

        self.jobTimeoutTimer.timeout.connect(functools.partial(abort, networkReply))
        self.jobTimeoutTimer.start(NetworkStateManager.MIN_JOB_INTERVAL - 500)

    def startTest(self):
        self.jobArrangeTimer.start(NetworkStateManager.MIN_JOB_INTERVAL)

    def stopTest(self):
        self.jobArrangeTimer.stop()

    def connectedCallback(self):
        self.jobInterval = NetworkStateManager.MIN_JOB_INTERVAL

        self.jobArrangeTimer.start(self.jobInterval)

    def disconnectedCallback(self):
        self.jobArrangeTimer.stop()



# Copyright (C) 2024  Loren Eteval <loren.eteval@proton.me>
#
# This file is part of Furious.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

from Furious.QtFramework import *
from Furious.QtFramework import gettext as _
from Furious.Utility import *

import functools

__all__ = ['RoutingAction']

BUILTIN_ROUTING = ['Bypass Mainland China', 'Global', 'Custom']

registerAppSettings('Routing', validRange=BUILTIN_ROUTING)

needTrans = functools.partial(needTransFn, source=__name__)

needTrans(*BUILTIN_ROUTING)


class RoutingChildAction(AppQAction):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def triggeredCallback(self, checked):
        textEnglish = self.textEnglish

        if AppSettings.get('Routing') != textEnglish:
            AppSettings.set('Routing', textEnglish)

            if APP().isSystemTrayConnected():
                APP().systemTray.ConnectAction.doDisconnect()
                APP().systemTray.ConnectAction.trigger()


needTrans('Routing')


class RoutingAction(AppQAction):
    def __init__(self, **kwargs):
        if AppSettings.get('Routing') == 'Bypass':
            # Update value for backward compatibility
            AppSettings.set('Routing', 'Bypass Mainland China')

        super().__init__(
            _('Routing'),
            icon=bootstrapIcon('shuffle.svg'),
            menu=AppQMenu(
                *list(
                    RoutingChildAction(
                        _(routing),
                        checkable=True,
                        checked=AppSettings.get('Routing') == routing,
                    )
                    for routing in BUILTIN_ROUTING
                ),
            ),
            **kwargs,
        )



# Copyright (C) 2024  Loren Eteval <loren.eteval@proton.me>
#
# This file is part of Furious.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

from __future__ import annotations

from abc import ABC

__all__ = ['UserServersTableItem']


class UserServersTableItem(ABC):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    @property
    def itemRemark(self) -> str:
        return ''

    @property
    def itemProtocol(self) -> str:
        return ''

    @property
    def itemAddress(self) -> str:
        return ''

    @property
    def itemPort(self) -> str:
        return ''

    @property
    def itemTransport(self) -> str:
        return ''

    @property
    def itemTLS(self) -> str:
        return ''

    @property
    def itemSubscription(self) -> str:
        return ''

    @property
    def itemLatency(self) -> str:
        return ''

    @property
    def itemSpeed(self) -> str:
        return ''



# Copyright (C) 2024  Loren Eteval <loren.eteval@proton.me>
#
# This file is part of Furious.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

from __future__ import annotations

from Furious.Interface import *
from Furious.QtFramework import *
from Furious.Library import *
from Furious.Utility import *

import time
import multiprocessing

__all__ = ['Tun2socks']


def startTun2socks(msgQueue: multiprocessing.Queue, *args):
    try:
        import tun2socks
    except ImportError:
        # Fake running process
        while True:
            time.sleep(1)
    else:
        if versionToValue(tun2socks.__version__) <= versionToValue('2.5.1.1'):
            redirect = False
        else:
            redirect = True

        StdoutRedirectHelper.launch(
            msgQueue, lambda: tun2socks.startFromArgs(*args), redirect
        )


class Tun2socks(CoreProcess):
    class ExitCode:
        # Windows shutting down
        SystemShuttingDown = 0x40010004

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

        self.cleanup = None

    @staticmethod
    def name() -> str:
        return 'Tun2socks'

    @staticmethod
    def version() -> str:
        try:
            import tun2socks

            return tun2socks.__version__
        except Exception:
            # Any non-exit exceptions

            return '0.0.0'

    def start(
        self,
        device: str,
        networkInterface: str,
        logLevel: str,
        proxy: str,
        restAPI: str,
        **kwargs,
    ) -> bool:
        return super().start(
            target=startTun2socks,
            args=(self.msgQueue, device, networkInterface, logLevel, proxy, restAPI),
            **kwargs,
        )

    def stop(self):
        SystemRoutingTable.deleteRelations()

        if callable(self.cleanup):
            self.cleanup()

        super().stop()
